#!/usr/bin/env python

__author__     = "Ekawat (Ice) Homsirikamol"
__copyright__  = "Copyright (c) 2015 CERG @ GMU, USA"
__license__    = "GPL 3.0"
__version__    = "1.0b1"

import types
import ctypes
import sys
import os
import inspect
from Crypto.Cipher import AES
from Crypto.Util import Counter
from Crypto.Util.number import long_to_bytes, bytes_to_long, size
# Note: Removed temporarily 
# from Crypto.Hash import CMAC


filepath = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
aes_chstone_path = filepath + os.pathsep + "AES_ECB/lib"
os.environ['PATH'] += os.pathsep + os.path.realpath(aes_chstone_path)

# from AES_GCM.aes_gcm import AES_GCM

def hexstr(val, blocksize=32):
    val = hex(val)
    if val[len(val)-1] == 'L':
        long = 1
    else:
        long = 0
    fill = (blocksize - (len(val)-(2+long))) % blocksize
    return val[2:len(val)-long].zfill(len(val)-(2+long)+fill).upper()



class InvalidInputException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


class InvalidTagException(Exception):
    def __str__(self):
        return 'The authenticaiton tag is invalid.'


# Galois/Counter Mode with AES-128 and 96-bit IV
class AES_ECB:
    def __init__(self, master_key):
        self.change_key(master_key)

    def change_key(self, master_key):
        self.aes_ecb = AES.new(master_key, mode=AES.MODE_ECB)

    def encrypt(self, plaintext):
        len_plaintext = len(plaintext)

        if len_plaintext > 0:
            if 0 != len_plaintext % 16:
                padded_plaintext = plaintext + \
                    b'\x00' * (16 - len_plaintext % 16)
            else:
                padded_plaintext = plaintext
            ciphertext = self.aes_ecb.encrypt(padded_plaintext)[:len_plaintext]
        else:
            ciphertext = b''

        return ciphertext

    def decrypt(self, ciphertext):
        len_ciphertext = len(ciphertext)
        if len_ciphertext > 0:
            if 0 != len_ciphertext % 16:
                padded_ciphertext = ciphertext + \
                    b'\x00' * (16 - len_ciphertext % 16)
            else:
                padded_ciphertext = ciphertext
            plaintext = self.aes_ecb.decrypt(padded_ciphertext)[:len_ciphertext]

        else:
            plaintext = b''

        return plaintext

# Removed temporarily
# class AES_CMAC:
    # def __init__(self, master_key):
        # self.change_key(master_key)

    # def change_key(self, master_key):
        # self.aes_cmac = CMAC.new(master_key, ciphermod=AES)

    # def digest(self, plaintext):
        # self.aes_cmac.update(plaintext)
        # digest = self.aes_cmac.digest()
        # return digest

class AES_CTR:
    def __init__(self, master_key, iv):
        self.master_key = master_key.decode("hex")
        self.iv = long(iv + "00000001", 16)

        if len(self.master_key) != 16:
            raise ValueError("Error: Invalid key size (Input = {0}). Must be 16!".format(len(self.master_key)))
        if len(iv.decode("hex")) != 12:
            raise ValueError("Error: Invalid IV size (Input = {0}). Must be 12!".format(len(iv.decode("hex"))))

    def encrypt(self, plaintext):
        ctr = Counter.new(128, initial_value=self.iv)
        self.aes_ctr = AES.new(self.master_key, AES.MODE_CTR, counter=ctr)

        len_plaintext = len(plaintext)
        if len_plaintext > 0:
            if 0 != len_plaintext % 16:
                padded_plaintext = plaintext + \
                    b'\x00' * (16 - len_plaintext % 16)
            else:
                padded_plaintext = plaintext
            ciphertext = self.aes_ctr.encrypt(padded_plaintext)[:len_plaintext]
        else:
            ciphertext = b''

        return ciphertext

    def decrypt(self, ciphertext):
        ctr_e = Counter.new(128, prefix=self.iv)
        self.aes_ctr = AES.new(self.master_key, AES.MODE_CTR, counter=ctr_e)

        len_ciphertext = len(ciphertext)
        if len_ciphertext > 0:
            if 0 != len_ciphertext % 16:
                padded_ciphertext = ciphertext + \
                    b'\x00' * (16 - len_ciphertext % 16)
            else:
                padded_ciphertext = ciphertext
            plaintext = self.aes_ctr.decrypt(padded_ciphertext)[:len_ciphertext]

        else:
            plaintext = b''

        return plaintext


class AES_HASH:
    def __init__(self):
        ### Initialize to all 1s in integer
        self.state      = (ctypes.c_int*32)()
        self.state_prev = (ctypes.c_int*32)()
        self.key   = (ctypes.c_int*32)()
        for i in range(0,32):
            self.state[i]      = 255
            self.state_prev[i] = 255
        self.mode  = ctypes.c_int(256256)

        self.aes_ecb = ctypes.cdll.aes_chstone
        self.aes_ecb.encrypt.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.c_int]
        self.aes_ecb.decrypt.argtypes = [ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.c_int]

    def digest_all(self, padded_text):
        blocks = len(padded_text)/64
        # Process blocks
        for i in range(0,blocks):
            # Get a block of data in hex
            block = padded_text[i*64:i*64+64]
            # turn into an array of integer to match the format required for aes_chstone
            for i in range(0,32):
                self.key[i] = int('0x'+block[i*2:i*2+2],16)
            # E_(x_i)[H_(i-1)]
            print('Data : {0}'.format(''.join("{0:02X}".format(self.state[i]) for i in range(0,32))))
            print('Key  : {0}'.format(''.join("{0:02X}".format(self.key[i]) for i in range(0,32))))
            self.aes_ecb.encrypt(self.state, self.key, self.mode)            
            print('Out  : {0}'.format(''.join("{0:02X}".format(self.state[i]) for i in range(0,32))))
            # E_(x_i)[H_(i-1)] ^ H_(i-1)
            for i in range(0,32):
                self.state[i]      = self.state[i] ^ self.state_prev[i]
                self.state_prev[i] = self.state[i]
        # Final Value
        # H = H_(n+1) = E_(H_n)[H_n] xor H_n
        print('Data : {0}'.format(''.join("{0:02X}".format(self.state[i]) for i in range(0,32))))
        print('Key  : {0}'.format(''.join("{0:02X}".format(self.state[i]) for i in range(0,32))))
        self.aes_ecb.encrypt(self.state, self.state, self.mode)
        print('Out  : {0}'.format(''.join("{0:02X}".format(self.state[i]) for i in range(0,32))))
        for i in range(0,32):
            self.state[i]      = self.state[i] ^ self.state_prev[i]
            self.state_prev[i] = self.state[i]

    def digest(self, plaintext):
        # Plaintext must be a HEX string , e.g. "12435667AF"
        len_plaintext = len(plaintext)
        bitsize       = (len_plaintext/2)*8

        # Perform padding to a multiple of 16-bytes
        if len_plaintext > 0:
            if 0 != len_plaintext % 32:
                padded_plaintext = plaintext + \
                    '0' * (32 - len_plaintext % 32)
            else:
                padded_plaintext = plaintext
        else:
            padded_plaintext = '0' * 32
        # Append input size and pad it to mutiple of 32-bytes
        len_padded = len(padded_plaintext)
        padded_plaintext = padded_plaintext + '{0:0{size}X}'.format(bitsize, size=(64-len_padded%64))

        print('\npadded plaintext {} \n'.format(padded_plaintext))
        # Digest the data
        self.digest_all(padded_plaintext)
        # Get output
        return ''.join("{0:02X}".format(self.state[i]) for i in range(0,32))

class AES_FORTUNA:
    def __init__(self, master_key, iv):
        self.aes_ctr = AES_CTR(master_key, iv)

    def generate(self, size):
        msg = "0"*(size*2)
        msg = msg.decode("hex")
        return self.aes_ctr.encrypt(msg)


#! This test is used in the original AES-GCM code in VHDL
def test_ecb_custom():
    master_key   = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
    plaintext    = b'\x00\x11\x22\x33\x44\x55\x66\x77\x88\x99\xaa\xbb\xcc\xdd\xee\xff'

    print('plaintext:', hexstr(bytes_to_long(plaintext)))

    ecb = AES_ECB(master_key)
    encrypted = ecb.encrypt(plaintext)
    print('encrypted:', hexstr(bytes_to_long(encrypted)))

#! Based on http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf F.1
def test_ecb1():
    master_key = 0x2b7e151628aed2a6abf7158809cf4f3c
    plaintext  = 0x6bc1bee22e409f96e93d7e117393172a

    plaintext = long_to_bytes(plaintext)
    master_key = long_to_bytes(master_key)
    print('plaintext:', hexstr(bytes_to_long(plaintext)))

    ecb = AES_ECB(master_key)
    encrypted = ecb.encrypt(plaintext)
    print('encrypted:', hexstr(bytes_to_long(encrypted)))


#! Based on http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf F.1
def test_ecb2():
    master_key = 0x2b7e151628aed2a6abf7158809cf4f3c
    ciphertext  = 0x3ad77bb40d7a3660a89ecaf32466ef97

    master_key = long_to_bytes(master_key)
    ciphertext = long_to_bytes(ciphertext)
    print('ciphertext:', hexstr(bytes_to_long(ciphertext)))

    ecb = AES_ECB(master_key)
    encrypted = ecb.decrypt(ciphertext)
    print('plaintext:', hexstr(bytes_to_long(encrypted)))

def test_cmac_rfc4493_ex1():
    master_key = 0x2b7e151628aed2a6abf7158809cf4f3c
    plaintext  = b''

    # plaintext = long_to_bytes(plaintext)
    master_key = long_to_bytes(master_key)
    print('plaintext:', hexstr(bytes_to_long(plaintext)))

    cmac = AES_CMAC(master_key)
    digest = cmac.digest(plaintext)
    print('digest:', hexstr(bytes_to_long(digest)))

def test_cmac_rfc4493_ex2():
    master_key = 0x2b7e151628aed2a6abf7158809cf4f3c
    plaintext  = 0x6bc1bee22e409f96e93d7e117393172a

    plaintext = long_to_bytes(plaintext)
    master_key = long_to_bytes(master_key)
    print('plaintext:', hexstr(bytes_to_long(plaintext)))

    cmac = AES_CMAC(master_key)
    digest = cmac.digest(plaintext)
    print('digest:', hexstr(bytes_to_long(digest)))

def test_cmac_rfc4493_ex3():
    master_key = 0x2b7e151628aed2a6abf7158809cf4f3c
    plaintext  = int("0x6bc1bee22e409f96e93d7e117393172aae2d8a571e03ac9c9eb76fac45af8e5130c81c46a35ce411", 16)

    plaintext = long_to_bytes(plaintext)
    master_key = long_to_bytes(master_key)
    print('plaintext:', hexstr(bytes_to_long(plaintext)))

    cmac = AES_CMAC(master_key)
    digest = cmac.digest(plaintext)
    print('digest:', hexstr(bytes_to_long(digest)))

def test_cmac_rfc4493_ex4():
    master_key = 0x2b7e151628aed2a6abf7158809cf4f3c
    plaintext  = int("0x6bc1bee22e409f96e93d7e117393172aae2d8a571e03ac9c9eb76fac45af8e5130c81c46a35ce411e5fbc1191a0a52eff69f2445df4f9b17ad2b417be66c3710", 16)

    plaintext = long_to_bytes(plaintext)
    master_key = long_to_bytes(master_key)
    print('plaintext:', hexstr(bytes_to_long(plaintext)))

    cmac = AES_CMAC(master_key)
    digest = cmac.digest(plaintext)
    print('digest:', hexstr(bytes_to_long(digest)))

def test_hash1():
    text = "00112233445566778899AABBCCDDEEFF"
    a = AES_HASH()
    print('digest:', a.digest(text))
def test_hash2():
    text = ""
    a = AES_HASH()
    print('digest:', a.digest(text))

# def test_gcm():
    # key  = "2b7e151628aed2a6abf7158809cf4f3c"
    # msg  = "6bc1bee22e409f96e93d7e117393172a"
    # iv   = "0B0B0B0B0B0B0B0B0B0B0B0B"
    # ad   = "6bc1bee22e409f96e93d7e117393172a"

    # key  = int('0x' + key,16)
    # iv   = int('0x' + iv, 16)
    # msg  = int('0x' + msg,16)
    # ad   = int('0x' + ad, 16)

    # msg = long_to_bytes(msg)
    # ad  = long_to_bytes(ad)
    # alg = AES_GCM(key)
    # (encrypted, tag) = alg.encrypt(iv, msg, ad)
    # encrypted = "{0:0{size}X}".format(bytes_to_long(encrypted), size=16*2)
    # tag       = "{0:0{size}X}".format(tag, size=16*2)

    # print('encrypted:',encrypted)
    # print('tag:',tag)


def test_ctr1():
    #based in http://tools.ietf.org/html/rfc3686
    key = "AE6852F8121067CC4BF7A5765577F39E" # replace this with a sensible value, preferably the output of a hash
    key = key.decode("hex")
    iv = "00000030000000000000000000000001" # replace this with a RANDOMLY GENERATED VALUE, and send this with the ciphertext!
    iv = long('0x' + iv, 16)
    plaintext = "Single block msg" # replace with your actual plaintext
    ctr = Counter.new(128, initial_value=long("00000030000000000000000000000001",16))
    cipher = AES.new(key, AES.MODE_CTR, counter=ctr)
    encrypted = cipher.encrypt(plaintext)
    print(encrypted.encode("hex").upper())

    #based on NIST 800-38a F.5.1 : http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf
    CTRkey="2b7e151628aed2a6abf7158809cf4f3c"
    CTRkey = long_to_bytes(long('0x' + CTRkey, 16))
    ctr=Counter.new(128, initial_value=int("f0f1f2f3f4f5f6f7f8f9fafbfcfdfeff",16))
    cipherCTR=AES.new(CTRkey, AES.MODE_CTR, counter=ctr)
    print(cipherCTR.encrypt("6bc1bee22e409f96e93d7e117393172a".decode("hex")).encode("hex").upper())

def test_ctr2():
    key = "7691BE035E5020A8AC6E618529F9A0DC"
    iv =  "00E0017B27777F3F4A1786F0"
    msg = "000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F20212223".decode("hex")
    alg = AES_CTR(key, iv)
    encrypted = alg.encrypt(msg)
    print('encrypted:',encrypted.encode("hex").upper())

def test_fortuna():
    key = "7E24067817FAE0D743D6CE1F32539163"
    iv =  "006CB6DBC0543B59DA48D90B"
    alg = AES_FORTUNA(key, iv)
    gen = alg.generate(1)
    print(gen.encode("hex"))

def test_hash3():
    text = "239988F5F4EB6C6DBC"
    a = AES_HASH()
    print('digest:', a.digest(text))
    
def test_custom():
    # CMAC
    master_key = 0x6B73B46CEEF01603CA919506B1B24E8C
    plaintext  = 0x4EF5B16002811B95A01F7E6F50F1A983AECDA5651B0F412EFA57F4ADB393F542A325618ACF5770BB4100126CB3E40000

    plaintext = long_to_bytes(plaintext)    
    master_key = long_to_bytes(master_key)
    print('plaintext:', hexstr(bytes_to_long(plaintext)))

    cmac = AES_CMAC(master_key)
    digest = cmac.digest(plaintext)
    print('digest:', hexstr(bytes_to_long(digest)))
    

if __name__ == '__main__':

    # test_ecb1()
    # print ''
    # test_ecb2()
    # test_ecb_custom()
    # print ''
    # test_ecb_custom()
    # print '\n==Test CMAC:'
    # test_cmac_rfc4493_ex1()
    # print ''
    # test_cmac_rfc4493_ex2()
    # print ''
    # test_cmac_rfc4493_ex3()
    # print ''
    # test_cmac_rfc4493_ex4()
    # print '\n==Test Hash:'
    # test_hash1()
    # print ''
    # test_hash2()
    # print ''
    # print '\n==Test AES-GCM:'
    # test_gcm()
    # print ''
    # print '\n==Test CTR:'
    # test_ctr1()

    # print '\ntest2:'
    # test_ctr2()
    # print '\n==Test AES-Fortuna:'
    # test_fortuna()

    print('\n==Test Custom:')
    test_hash3()