#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
AETVGen.py: Generate test vectors for the Authenticated Encryption algorithms according
            to protocol specified in.

Syntax Note:  "_WIDTH" always have the unit of bit(s)
              "_SIZE"  always have the unit of byte(s)
              "_WORD"  refers to the number of WORDS for a given widthData
"""

__author__     = "Ekawat (Ice) Homsirikamol"
__copyright__  = "Copyright (c) 2015 CERG @ GMU, USA"
__license__    = "GPL 3.0"
__version__    = "1.0b1"

# Core imports
import random
import sys
import os
import binascii
# AES
import AES

# ===================================
# CFFI - External library declaration
# ===================================
import cffi
ffi = cffi.FFI()
ffi.cdef("""
    int crypto_aead_encrypt(
        unsigned char *c,unsigned long long *clen,
        const unsigned char *m,unsigned long long mlen,
        const unsigned char *ad,unsigned long long adlen,
        const unsigned char *nsec,
        const unsigned char *npub,
        const unsigned char *k
	);
    int crypto_aead_decrypt(
        unsigned char *m,unsigned long long *mlen,
        unsigned char *nsec,
        const unsigned char *c,unsigned long long clen,
        const unsigned char *ad,unsigned long long adlen,
        const unsigned char *npub,
        const unsigned char *k
	);
""")

# ====================
# Algorithms and paths
# ====================

LIB_DIR = os.path.realpath("../lib")
if os.name == "nt":
    # windows
    try:
        os.environ['path'] += os.pathsep + os.path.realpath("../lib")
    except KeyError:
        os.environ['path'] = os.path.realpath("../lib")

else:
    # linux
    # This doesn't actually do anything, have to set it in the shell *before*
    # Correct way would be to use ctypes.cdll.LoadLibrary() with the full path,
    # running the script.
    # including extension in lines 264-440
    # XXX: Ice, fix this. I'm too lazy to
    try:
        os.environ['LD_LIBRARY_PATH'] += os.pathsep + os.path.realpath("../lib")
    except KeyError:
        os.environ['LD_LIBRARY_PATH'] = os.path.realpath("../lib")

# Supported imports
from Crypto.Util.number import long_to_bytes, bytes_to_long, size

class AETVGenError(Exception):
    """Class of error used in the AE implementation
    Use: raise AETVGenError.AETVGenError("Text to be displayed")"""
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class AETVgen:
    """
    Class for generating test vector for Keccak
    """

    # Constant declarations
    INS_PREFIX = 'INS'
    HDR_PREFIX = 'HDR'
    DAT_PREFIX = 'DAT'

    OPCODE_ENCRYPT      = 0
    OPCODE_DECRYPT      = 1
    OPCODE_AEAD_ENCRYPT = 2
    OPCODE_AEAD_DECRYPT = 3
    OPCODE_LOAD_KEY     = 4
    OPCODE_ACT_KEY      = 5
    OPCODE_MAC          = 6
    OPCODE_HASH         = 7
    OPCODE_PRNG         = 8
    CODE_OPCODE = {OPCODE_ENCRYPT:'Encrypt', OPCODE_DECRYPT:'Decrypt', OPCODE_AEAD_ENCRYPT:'AEAD encrypt', OPCODE_AEAD_DECRYPT:'AEAD decrypt',
                   OPCODE_LOAD_KEY:'Load key', OPCODE_ACT_KEY:'Activate key', OPCODE_MAC:'MAC', OPCODE_HASH:'HASH', OPCODE_PRNG:'PRNG'}

    STYPE_RESERVED      = 0
    STYPE_NPUB          = 1
    STYPE_AD            = 2
    STYPE_PLAINTEXT     = 3
    STYPE_CIPHERTEXT    = 4
    STYPE_TAG           = 5
    STYPE_KEY           = 6
    STYPE_NSEC          = 7
    STYPE_NSEC_CIPH     = 8
    CODE_STYPE = {STYPE_RESERVED:'Reserved', STYPE_NPUB:'NPUB', STYPE_AD:'Associated Data', STYPE_PLAINTEXT:'Message',
                  STYPE_CIPHERTEXT:'Ciphertext', STYPE_TAG:'Tag', STYPE_KEY:'Key', STYPE_NSEC:'NSEC', STYPE_NSEC_CIPH:'NSEC Cipher'}

    FILE_ALL            = 0
    FILE_INPUT          = 1
    FILE_OUTPUT         = 2
    FILE_PUBLIC         = 3
    FILE_PUBLIC_INPUT   = 4
    FILE_SECRET         = 5

    def __init__(self, algorithm, **kwargs):
        self.param = dict()
        # Required values
        self.param['algorithm']      = algorithm
        if 'icepole' in self.param['algorithm']:
            self.param['enableNSEC'] = True

        # Optional settings
        self.param['stdout']         = False
        self.param['testMode']       = None     # Mode for fixing the message and key value to one block
        self.param['debugMsgMode']   = 0        # 0 = all 0xFF, else = step (0102...)
        self.param['libVer']         = None     # Library version (Unused)
        self.param['sizeKey']        = 16
        self.param['sizeNPUB']       = 12
        self.param['sizeTag']        = 16
        self.param['minSizeAD']      = 1
        self.param['maxSizeAD']      = 512
        self.param['minSizeM']       = 1
        self.param['maxSizeM']       = 512
        self.param['minSizeRNG']     = 1
        self.param['maxSizeRNG']     = 512
        self.param['filePDI']        = 'pdi.txt'
        self.param['fileSDI']        = 'sdi.txt'
        self.param['fileDO']         = 'do.txt'
        self.param['sizePIO']        = 16
        self.param['sizeSIO']        = 16
        self.param['maxSizeSegment'] = 100000
        self.param['opcode']         = False
        self.param['sizeNSEC']       = 16
        self.param['enableNSEC']     = False
        self.param['verbose']        = False
        self.param['decrypt']        = False    # Perform decryption after encryption (Only activated when 'verbose' is True)
        self.ABYTES                  = 100      # Default additional bytes (Used for memory allocation)
        self.reverseDecrypt          = False    # Reverse block order for decryption
        self.plainTextMode           = 0        # Default write mode (1 = No NPUB segment with AD = AD + NPUB)
        self.cipherTextMode          = 0        # Default write mode (1 = No NPUB segment with AD = AD + NPUB)
        """
        PLAINTEXT_MODE:
        {0} N_A_M: Separate Nonce, Associated Data, and Message segments.
        {1} NA_M: The Associated Data segment contains Nonce concatenated with Associated Data.
        {2} AN_M: The Associated Data segment contains Associated Data concatenated with Nonce.
        {3} N_A_M_A: Separate Nonce, Associated Data - Header, Message, and Associated Data - Trailer segments.

        CIPHERTEXT_MODE:
        {0} C_T: Separate Ciphertext and Tag segments.
        {1} CT : The Ciphertext segment contains Ciphertext concatenated with Tag.
        {2} Cexp_T: Separate expanded Ciphertext

        When
            CIPHERTEXT_MODE = CT
        then
        1) during encryption
            a. len_d = |M|
            b. The tag output of the Datapath is not used
            c. The output processor does not wait for 1 at the tag_write output of the Datapath
            d. The size of C in the ciphertext segment header = |M|+|T|

        2) during decryption
            a. The size of C in the ciphertext segment header = |M|+|T|
            b. len_d = |M|+|T| (|M| is calculated inside of the datapath)
            c. The exp_tag input of the Datapath is not used.
            d. The exp_tag_ready input of the Controller is not used.

        When
            CIPHERTEXT_MODE = C_T
        then
        1) during encryption
            a. len_d = |M|
            b. The tag output of the Datapath is used
            c. The output processor waits for 1 at the tag_write output of the Datapath
            d. The size of C in the ciphertext segment header = |M|

        2) during decryption
            a. The size of C in the ciphertext segment header = |M|
            b. len_d = |M|
            c. The exp_tag input of the Datapath is used.
            d. The exp_tag_ready input of the Controller is used.

        When
            CIPHERTEXT_MODE = Cexp_T
        then
        1) during encryption
            a. len_d = |M|
            b. The tag output of the Datapath is used
            c. The output processor waits for 1 at the tag_write output of the Datapath
            d. The size of C in the ciphertext segment header = |M| but the output processor expects n*ceil(|M|/n) bits of the ciphertext

        2) during decryption
            a. The size of C in the ciphertext segment header = |M|,
               but the input processor reads and passes to the Datapath n*ceil(|M|/n) bits of the ciphertext
            b. len_d = |M|
            c. The exp_tag input of the Datapath is used.
            d. The exp_tag_ready input of the Controller is used.
        """

        for key, value in kwargs.items():
            if key.lower()              == 'stdout'.lower():
                self.param['stdout']     = value
            elif key.lower()              == 'testMode'.lower():
                if value not in [0, 1, None]:
                    raise AETVGenError.AETVGenError("Invalid 'testMode' option: {}".format(value))
                self.param['testMode'] = value
            elif key.lower()              == 'debugMsgMode'.lower():
                self.param['debugMsgMode'] = value
            elif key.lower()              == 'decrypt'.lower():
                self.param['decrypt'] = value
            elif key.lower()            == 'sizeKey'.lower():
                self.param['sizeKey']   = value
            elif key.lower()            == 'sizeNPUB'.lower():
                self.param['sizeNPUB']    = value
            elif key.lower()            == 'sizeTag'.lower():
                self.param['sizeTag']   = value
            elif key.lower()            == 'minSizeAD'.lower():
                self.param['minSizeAD'] = value
            elif key.lower()            == 'maxSizeAD'.lower():
                self.param['maxSizeAD'] = value
            elif key.lower()            == 'minSizeM'.lower():
                self.param['minSizeM']  = value
            elif key.lower()            == 'maxSizeM'.lower():
                self.param['maxSizeM']  = value
            elif key.lower()            == 'maxSizeSegment'.lower():
                self.param['maxSizeSegment'] = value
            elif key.lower()            == 'sizePIO'.lower():
                self.param['sizePIO'] = value
            elif key.lower()            == 'sizeSIO'.lower():
                self.param['sizeSIO'] = value
            elif key.lower()            == 'opcode'.lower():
                self.param['opcode'] = value
            elif key.lower()            == 'verbose'.lower():
                self.param['verbose'] = value
            elif key.lower()            == 'enableNSEC'.lower():
                if value == 1 or value == True:
                    self.param['enableNSEC'] = True
                else:
                    self.param['enableNSEC'] = False
            else:
                raise AETVGenError.AETVGenError("Invalid init argument, unknown key -> " + key)

        if (self.param['maxSizeSegment'] % self.param['sizePIO'] > 0) or \
           (self.param['maxSizeSegment'] % self.param['sizeSIO'] > 0):
            raise ValueError("Maximum segment size has to be a multiple of I/O width")

        if self.param['enableNSEC'] == False:
            self.param['sizeNSEC'] = 0

        # Default settings
        self.pdi   = ''
        self.pdo   = ''
        self.do    = ''
        self.msgid = 0
        self.keyid = 0
        self.key   = ''
        self.blockSize = 16
        (expTagSize, expKeySize, expNPUBSize) = (0, 0, 0)
        lib        = "aescopav1"

        self.supportedCaesarAlg = ["ACORN", "AES-COPA", "AES-GCM", "CLOC", "ICEPOLE", "Joltik",
            "Ketje-Jr", "Keyak", "Minalpher", "Poet", "PAEQ",
            "PRIMATES-APE", "PRIMATES-HANUMAN", "PRIMATES-GIBBON",
            "Scream"]

        #### Functions specific checks and declaration for CAESAR Algorithms
        if algorithm.lower() in [a.lower() for a in self.supportedCaesarAlg]:
            if algorithm.lower() == "acorn":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 16
                lib                 = "acorn128"
                self.blockSize      = 16
            elif algorithm.lower() == "aes-copa":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 16
                lib                 = "aescopav1"
                self.blockSize      = 16
            elif algorithm.lower() == "aes-gcm":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 12
                lib                 = "aes128gcmv1"
                self.blockSize      = 16
            elif algorithm.lower() == "cloc":
                expTagSize          = 8
                expKeySize          = 16
                expNPUBSize         = 12
                lib                 = "aes128n12clocv1"
                self.blockSize      = 16
            elif algorithm.lower() == "icepole":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 12
                lib                 = "icepole128av1"
                self.blockSize      = 1024/8
            elif algorithm.lower() == "joltik":
                expTagSize          = 8
                expKeySize          = 16
                expNPUBSize         = 4
                lib                 = "joltikneq12864v1"
                self.blockSize      = 8
                #PRIMATEs
            elif algorithm.lower() == "ketje-jr":    # Ketje Jr
                expTagSize          = 8
                expKeySize          = 16
                expNPUBSize         = 16
                lib                 = "ketjejrv1"
                self.blockSize      = 200/8
            elif algorithm.lower() in ["lakekeyak", "keyak", "caesar-keyak"]:    # Keyak
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 16
                lib                 = "lakekeyakv1"
                self.blockSize      = 1344/8
            elif algorithm.lower() == "minalpher":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 13
                self.blockSize      = 256/8
                self.cipherTextMode = 2
                lib                 = "minalpherv1"
            elif algorithm.lower() == "paeq":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 12
                self.blockSize      = 46
                lib                 = "paeq128"
            elif algorithm.lower() == "poet":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 16
                self.blockSize      = 16
                lib                 = "paeq128"
            elif algorithm.lower() in ["primates-ape","ape"]:        # APE-120 (PRIMATEs-120)
                expTagSize          = 15*2
                expKeySize          = 15*2
                expNPUBSize         = 15
                lib                 = "primatesv1ape120"
                self.blockSize      = 5
                self.reverseDecrypt = True
                self.cipherTextMode = 2
            elif algorithm.lower() in ["primates-hanuman","hanuman"]:    # HANUMAN-120 (PRIMATEs-120)
                expTagSize          = 15
                expKeySize          = 15
                expNPUBSize         = 15
                lib                 = "primatesv1hanuman120"
                self.blockSize      = 5
            elif algorithm.lower() in ["primates-gibbon","gibbon"]:     # GIBBON-120 (PRIMATEs-120)
                expTagSize          = 15
                expKeySize          = 15
                expNPUBSize         = 15
                lib                 = "primatesv1gibbon120"
                self.blockSize      = 5
            elif algorithm.lower() == "scream":
                expTagSize          = 16
                expKeySize          = 16
                expNPUBSize         = 12
                lib                 = "Scream10v2"
                self.blockSize      = 16
            else:
                print ("Warning: Algorithm {} not found".format(self.algorithm))

            # ============
            # Error checks
            # ============
            if (expTagSize > 0 and expTagSize != self.param['sizeTag']):
                raise ValueError("{}: Tag size must be {}-bit".format(algorithm, expTagSize))
            if (expKeySize > 0 and expKeySize != self.param['sizeKey']):
                raise ValueError("{}: Key size must be {}-bit".format(algorithm, sizeKey))
            if (expNPUBSize> 0 and expNPUBSize!= self.param['sizeNPUB']):
                print(expNPUBSize)
                raise ValueError("{}: NPUB  size must be {}-bit".format(algorithm, expTagSize))

            # ================
            # CAESAR library declaration
            # ================
            if self.param['verbose'] == True:
                lib    = "{}_verbose".format(lib)
            if os.name == "nt":
                lib    = "{}.dll".format(lib)
            else:
                lib    = "{}/{}.so".format(os.path.realpath("../lib"), lib)
            self.caesar = ffi.dlopen(lib)
        else:
            raise ValueError("Unsupported CAESAR Algorithm: {}.\n Supported algorithms are {}".format(algorithm, self.supportedCaesarAlg))



    def getHdrIns(self, opcode, iowidth):
        """
        Format of instruction header is as followed:
        8-bit   message ID
        4-bit   zeros
        4-bit   opcode
        8-bit   key ID
        others  zeros
        """
        # Set lower bound for header width
        if iowidth < 4:
            iowidth = 4
        ins    = "{0:02X}{z:01X}{1:01X}{2:02X}{z:{z}{fill}X}".format(self.msgid,opcode,self.keyid, z=0, fill=self._calcIOFill(6, iowidth))
        return ins


    def getHdrCmd(self, segType, pad, eoi, eot, size, iowidth):
        """
        Format of command header is as followed:
        8-bit   message ID
        4-bit   segment type
        1-bit   reserved
        1-bit   padding flag
        1-bit   last segment of message flag
        1-bit   last segment of segment type flag
        others  size
        """
        # Set lower bound for header width
        if iowidth < 4:
            iowidth = 4
        nibble = int("0{0}{1}{2}".format(pad,eoi,eot), 2)
        hdr    = "{0:02X}{1:01X}{2:01X}{3:{z}{fill}X}".format(self.msgid,segType,nibble,int(size), z=0, fill=self._calcIOFill(4, iowidth))
        return hdr


    def compute(self, algorithm, opcode, key, npub, ad, sizeAD, msg, sizeMsg, nsec):
        (encrypted, sizeEncrypted, tag, nsec_ciph) = ('',sizeMsg,'','')
        # OPCODE check
        if   opcode == self.OPCODE_ENCRYPT or opcode == self.OPCODE_AEAD_ENCRYPT or \
             opcode == self.OPCODE_DECRYPT or opcode == self.OPCODE_AEAD_DECRYPT:
            if algorithm.lower() not in [a.lower() for a in self.supportedCaesarAlg]:
                raise ValueError("Invalid OPCODE of operation for the specific algorithm : {0}".format(algorithm))
        if algorithm.lower() in ['aes_cmac', 'aes-cmac']:
            # key = int('0x' + key,16)
            # msg = int('0x' + msg,16)
            # key = long_to_bytes(key)
            # msg = long_to_bytes(msg)
            alg = aes.AES_CMAC(key)
            tag = alg.digest(msg)
            tag = "{0:0{size}X}".format(bytes_to_long(tag), size=self.param['sizeTag']*2)
        elif algorithm.lower() in ['aes_ctr', 'aes-ctr']:
            alg = aes.AES_CTR(key, npub)
            encrypted = alg.encrypt(msg.decode("hex"))
            encrypted = encrypted.encode("hex").upper()
        elif algorithm.lower() in ['aes_ecb', 'aes-ecb']:
            key  = int('0x' + key,16)
            msg  = int('0x' + msg,16)
            key = long_to_bytes(key)
            msg = long_to_bytes(msg)
            alg = AES.AES_ECB(key)
            encrypted = alg.encrypt(msg)
            encrypted = "{0:0{size}X}".format(bytes_to_long(encrypted), size=sizeMsg*2)
        elif algorithm.lower() in ['aes_fortuna', 'aes-fortuna']:
            alg = AES.AES_FORTUNA(key, npub)
            encrypted = alg.generate(sizeMsg).encode("hex").upper()
        elif algorithm.lower() in ['aes_hash', 'aes-hash']:
            alg = AES.AES_HASH()
            tag = alg.digest(msg)
        elif    algorithm.lower() in ['aes_ocb3', 'aes-ocb3']:
            key  = binascii.unhexlify(key)
            npub = binascii.unhexlify(npub)
            ad   = binascii.unhexlify(ad)
            msg  = binascii.unhexlify(msg)
            # Allocate space for output
            c    = "0"*sizeMsg + "0"*self.param['sizeTag']
            OCB_ENCRYPT = 1
            OCB_DECRYPT = 0
            self.aes_ocb3.ocb_crypt(c, key, npub, ad, sizeAD, msg, sizeMsg, OCB_ENCRYPT)
            nsec_ciph = nsec    # Unused
            encrypted = "".join(c[i].encode('hex').upper() for i in range(0,sizeMsg))
            tag       = "".join(c[i].encode('hex').upper() for i in range(sizeMsg,sizeMsg+self.param['sizeTag']))
        elif algorithm.lower() in [a.lower() for a in self.supportedCaesarAlg]:   # CAESAR Interface
            ### Input
            m       = ffi.new("char[]", binascii.unhexlify(msg))
            mlen    = ffi.cast("unsigned long long", sizeMsg)
            c       = ffi.new("char[]", sizeMsg+self.ABYTES)
            clen    = ffi.new("unsigned long long *", sizeMsg+self.ABYTES)
            ad      = ffi.new("char[]", binascii.unhexlify(ad))
            adlen   = ffi.cast("unsigned long long", sizeAD)
            nsec    = ffi.new("char[]", binascii.unhexlify(nsec))
            npub    = ffi.new("char[]", binascii.unhexlify(npub))
            k       = ffi.new("char[]", binascii.unhexlify(key))

            ### ABI leve, in-line call
            self.caesar.crypto_aead_encrypt(c, clen, m, mlen, ad, adlen, nsec, npub, k)

            ### Output
            # Note: Cannot use ffi.string() as it will truncate at null (0x00)
            output = "".join(binascii.hexlify(c[i]).decode("utf-8").upper() for i in range(0, clen[0]))

            encrypted     = output[:int((clen[0]-self.param['sizeTag'])*2)]
            sizeEncrypted = clen[0]-self.param['sizeTag']
            tag           = output[int((clen[0]-self.param['sizeTag'])*2):]
            nsec_ciph     = binascii.hexlify(ffi.string(nsec,self.param['sizeNSEC'])).decode("utf-8").upper()

            ### Decryption
            if (self.param['decrypt'] == True):
                if sizeMsg > 0:
                    print("\n\t===========================")
                    print("\t        Decrypt")
                    print("\t===========================\n")

                    mlen    = ffi.new("unsigned long long *", 1)
                    clen    = ffi.cast("unsigned long long", clen[0])
                    self.caesar.crypto_aead_decrypt( m, mlen, nsec, c, clen, ad, adlen, npub, k)
                else:
                    print("No decryption! Message size = 0")

        elif algorithm.lower() in ['keccak-hash', 'sha3']:
            alg = Keccak.Keccak()
            msg = (sizeMsg*8, msg)
            (tag, _) = alg.Hash(msg, n=256, r=1088, c=512,verbose=False)
        elif 'keccak-mac'       == algorithm.lower():
            alg = Keccak.Keccak()
            msg = (sizeMsg*8, msg)
            key = (self.param['sizeKey']*8, key)
            (tag, _) = alg.MAC(key, msg, n=256, r=1088, c=512,verbose=False)
        elif 'keccak-prng'      == algorithm.lower():
            alg = Keccak.Keccak()
            seed = (self.param['sizeNPUB']*8, npub)
            (encrypted, _) = alg.PRNG(seed,n=512,r=1088,c=512,verbose=False)
        elif algorithm.lower() in ['lakekeyak', 'keccak-ae', 'keccak-keyak']:
            alg = Keccak.Keccak()
            M   = (sizeMsg*8, msg)
            N   = (self.param['sizeNPUB']*8, npub)
            A   = (sizeAD*8, ad)
            K   = (self.param['sizeKey']*8, key)
            ([encrypted, tag], _) = alg.LakeKeyak(K, N, A, M,  n=128,r=1348,c=252,enc=1,verbose=True)
        else:
            raise ValueError("Unsupported algorithm : {0}.".format(algorithm))

        return (encrypted, sizeEncrypted, tag, nsec_ciph)


    def genTV(self, algorithm, opcode, sizeNPUB, sizeAD, sizeMsg, newKey):
        if sizeMsg == 0 and opcode in [self.OPCODE_AEAD_DECRYPT]:
            print("Warning! Message of size 0 is not currently supported for authenticated")
            print("decryption operation by our Pre- and Post-processors (v1.0b1).")            
        """ Generate a single test vector """
        self.msgid = self.msgid + 1
        ### Message header information
        text = "Ciph" if opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT] else "Msg"
        self._pp("\n#### MsgID={0:03}, KeyID={1:03}, AD Size={2: 4}, {6} Size={3: 4}, Opcode={4}, Algorithm={5}".format(self.msgid, self.keyid, sizeAD, sizeMsg, self.CODE_OPCODE[opcode], algorithm, text), self.FILE_PUBLIC)

        ### Generate New Key, if required.
        if (newKey == 1 or self.key == '') and \
           (opcode in \
                [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, \
                 self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, \
                 self.OPCODE_MAC, self.OPCODE_PRNG]):
            self.keyid = self.keyid + 1
            newKey = 1
            if self.param['testMode'] == 0:
                self.key = "55"*self.param['sizeKey']
            else:
                self.key = self._genRandomData(self.param['sizeKey'])
            # SDI
            (pad, eoi, eot) = (0, 1, 1)
            self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, self.CODE_OPCODE[self.OPCODE_LOAD_KEY]), self.FILE_SECRET)
            self._pp("INS = {0}".format(self.getHdrIns(self.OPCODE_LOAD_KEY,self.param['sizeSIO'])), self.FILE_SECRET)
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_KEY], self.param['sizeKey'], pad, eoi, eot), self.FILE_SECRET)
            self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_KEY, pad, eoi, eot, self.param['sizeKey'], self.param['sizeSIO'])), self.FILE_SECRET)
            self._pp("DAT = {0}{1}".format(self.key, '0'*self._calcIOFill(self.param['sizeKey']*2,self.param['sizeSIO'])), self.FILE_SECRET)
            # PDI, DO
            self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, self.CODE_OPCODE[self.OPCODE_ACT_KEY]), self.FILE_PUBLIC)
            self._pp("INS = {0}".format(self.getHdrIns(self.OPCODE_ACT_KEY,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)

        ### Generate input data
        nsec = "00000000000000000000000000000000"
        if self.param['testMode'] == 0:
            npub   = "B0"*sizeNPUB
            ad   = "A0"*sizeAD
            if self.param['debugMsgMode'] == 0:
                msg  = "FF"*sizeMsg
            else:
                msg = "".join("{0:02X}".format(j%256) for j in range(1,sizeMsg+1))
        else:
            if (sizeNPUB > 0):
                npub = self._genRandomData(sizeNPUB)
            else:
                npub = ""
            ad   = self._genRandomData(sizeAD)
            msg  = self._genRandomData(sizeMsg)
            if self.param['enableNSEC'] == True:
                nsec = self._genRandomData(self.param['sizeNSEC'])

        (pad, eoi, eot) = (0, 0, 0)

        ### Calculate Output data
        (encrypted, sizeEncrypted, tag, nsec_ciph) = self.compute(algorithm, opcode, self.key, npub, ad, sizeAD, msg, sizeMsg, nsec)
        sizeInput = sizeMsg
        sizeOutput = sizeEncrypted

        ### Swap msg and encrypted data if decryption
        if opcode in [self.OPCODE_DECRYPT, self.OPCODE_AEAD_DECRYPT]:
            ### Reverse ordering can only occurs for cipherTextExpansion mode

            if ((self.cipherTextMode == 2) and (self.reverseDecrypt == True)):
                fullBlockLen = (len(msg)/(self.blockSize*2))*(self.blockSize*2)
                ### Pad message with X
                if (len(msg) % self.blockSize*2 != 0):
                    msg = msg[:fullBlockLen] + self._xpad(msg[fullBlockLen:], self.blockSize)

                ### Reverse order
                newMsg       = ""
                newEncrypted = ""
                for i in reversed(range(0, len(encrypted)/(self.blockSize*2))):
                    newEncrypted = newEncrypted + encrypted[i*(self.blockSize*2):(i+1)*(self.blockSize*2)]
                    newMsg       = newMsg       +       msg[i*(self.blockSize*2):(i+1)*(self.blockSize*2)]
                encrypted = newEncrypted.replace("X","0")
                msg       = newMsg

            (encrypted, msg)        = (msg, encrypted)
            (sizeInput, sizeOutput) = (sizeOutput, sizeInput)
            (nsec_ciph, nsec)       = (nsec, nsec_ciph)
            ### Error checking
            if sizeOutput != sizeInput:
                if algorithm not in ['ape','minalpher']:
                    self._printCallInfo(algorithm, opcode, self.key, npub, ad, msg, encrypted, tag)
                    raise ValueError("Error! encrypted output length is not the same as the input message length.")

        ### Instruction
        self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, self.CODE_OPCODE[opcode]), self.FILE_PUBLIC_INPUT)
        self._pp("INS = {0}".format(self.getHdrIns(opcode,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
        if opcode not in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]:  # No instruction for output
            self._pp("# {0:03} : Instruction (Opcode={1})".format(self.msgid, self.CODE_OPCODE[opcode+1]), self.FILE_OUTPUT)
            self._pp("INS = {0}".format(self.getHdrIns(opcode+1,self.param['sizePIO'])), self.FILE_OUTPUT)


        ### NPUB
        if (sizeNPUB > 0) and (self.plainTextMode == 0) and \
           (opcode in [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, \
                       self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, \
                       self.OPCODE_PRNG]):
            eot = 1
            # Input PDI file
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_NPUB],self.param['sizeNPUB'],pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_NPUB,pad,eoi,eot,self.param['sizeNPUB'],self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("DAT = {0}".format(self._zpad(npub,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            # Output DO file (remove if AEAD_DECRYPT)
            if opcode not in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]:
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_NPUB],self.param['sizeNPUB'],pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_NPUB,pad,eoi,eot,self.param['sizeNPUB'],self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(self._zpad(npub,self.param['sizePIO'])), self.FILE_OUTPUT)
        elif (self.plainTextMode == 1): # NA_M
            self._pp("# NPUB is appended to AD", self.FILE_PUBLIC_INPUT)
            self._pp("# NPUB = {0}".format(npub), self.FILE_PUBLIC)
            ad = npub + ad
            sizeAD = sizeAD + self.param['sizeNPUB']
        elif (self.plainTextMode == 2): # AN_M
            self._pp("# NPUB is appended to AD", self.FILE_PUBLIC_INPUT)
            self._pp("# NPUB = {0}".format(npub), self.FILE_PUBLIC)
            ad = ad + npub
            sizeAD = sizeAD + self.param['sizeNPUB']


        ### NSEC
        if (self.param['enableNSEC'] == True) and \
           (opcode in [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT]):
            if opcode == self.OPCODE_DECRYPT or opcode == self.OPCODE_AEAD_DECRYPT:
                code_nsec_in  = self.STYPE_NSEC_CIPH
                code_nsec_out = self.STYPE_NSEC
            else:
                code_nsec_in  = self.STYPE_NSEC
                code_nsec_out = self.STYPE_NSEC_CIPH
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[code_nsec_in], self.param['sizeNSEC'],pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(code_nsec_in,pad,eoi,eot,self.param['sizeNSEC'],self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("DAT = {0}".format(self._zpad(nsec,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            # Remove NSEC
            if opcode not in [self.OPCODE_AEAD_DECRYPT, self.OPCODE_DECRYPT]:
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[code_nsec_out],self.param['sizeNSEC'],pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(code_nsec_out,pad,eoi,eot,self.param['sizeNSEC'],self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(self._xpad(nsec_ciph,self.param['sizePIO'])), self.FILE_OUTPUT)


        if ((opcode == self.OPCODE_AEAD_DECRYPT) and (self.reverseDecrypt == True) and (self.cipherTextMode == 2)):
            (pad, eoi, eot) = (0, 0, 1)
            self._genTagSegment(tag, opcode, pad, eoi, eot)

        ### AD
        if opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT]:
            eot             = 0
            fullSegment     = int(sizeAD/self.param['maxSizeSegment'])
            sgtSize         = self.param['maxSizeSegment']
            finalSgtSize    = sizeAD % self.param['maxSizeSegment']

            #   Full blocks
            for i in range(0,fullSegment):
                if i == fullSegment-1 and finalSgtSize == 0:
                    eot = 1
                    if sizeInput > 0:
                        self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],sgtSize,pad,eoi,eot), self.FILE_PUBLIC)
                        self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,sgtSize,self.param['sizePIO'])), self.FILE_PUBLIC)
                    else:
                         # AD (input)
                        eoi = 0 if opcode in [self.OPCODE_AEAD_DECRYPT] else 1
                        self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],sgtSize,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                        self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,sgtSize,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                        # AD (output)
                        eoi = 0 if opcode not in [self.OPCODE_AEAD_DECRYPT] else 1
                        self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],sgtSize,pad,eoi,eot), self.FILE_SECRET)
                        self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,sgtSize,self.param['sizePIO'])), self.FILE_OUTPUT)
                    self._pp("DAT = {0}".format(ad[i*self.param['maxSizeSegment']*2:(i+1)*self.param['maxSizeSegment']*2]), self.FILE_PUBLIC)
                else:
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],sgtSize,pad,eoi,eot), self.FILE_PUBLIC)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,sgtSize,self.param['sizePIO'])), self.FILE_PUBLIC)
                    self._pp("DAT = {0}".format(ad[i*self.param['maxSizeSegment']*2:(i+1)*self.param['maxSizeSegment']*2]), self.FILE_PUBLIC)
            #   Partial block
            if finalSgtSize > 0:
                eot = 1
                if sizeInput > 0:
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],finalSgtSize,pad,eoi,eot), self.FILE_PUBLIC)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_PUBLIC)
                else:
                    # AD (input)
                    eoi = 0 if opcode in [self.OPCODE_AEAD_DECRYPT] else 1
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],finalSgtSize,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                    # AD (output)
                    eoi = 0 if opcode not in [self.OPCODE_AEAD_DECRYPT] else 1
                    self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_AD],finalSgtSize,pad,eoi,eot), self.FILE_OUTPUT)
                    self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_AD,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(self._zpad(ad[int(fullSegment*self.param['maxSizeSegment']*2):],self.param['sizePIO'])), self.FILE_PUBLIC)

        ### Data (Input)
        if (sizeInput > 0) and (opcode not in [self.OPCODE_PRNG]):
            if opcode == self.OPCODE_DECRYPT or opcode == self.OPCODE_AEAD_DECRYPT:
                itype = self.STYPE_CIPHERTEXT
            else:
                itype = self.STYPE_PLAINTEXT

            eot             = 0
            fullSegment     = int(sizeInput/self.param['maxSizeSegment'])
            sgtSize         = self.param['maxSizeSegment']
            finalSgtSize    = sizeInput % self.param['maxSizeSegment']
            # Full blocks
            for i in range(0,fullSegment):
                if i == fullSegment-1 and finalSgtSize == 0:
                    eot = 1
                    # Encryption has no more segment (no tag)
                    if opcode == self.OPCODE_AEAD_ENCRYPT:
                        eoi = 1
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[itype],sgtSize,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(itype,pad,eoi,eot,sgtSize,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                self._pp("DAT = {0}".format(msg[i*self.param['maxSizeSegment']*2:(i+1)*self.param['maxSizeSegment']*2]), self.FILE_PUBLIC_INPUT)

            # Partial blocks
            if finalSgtSize > 0:
                eot = 1
                # Encryption has no more segment (no tag)
                if (opcode == self.OPCODE_AEAD_ENCRYPT) or \
                   (opcode == self.OPCODE_AEAD_DECRYPT and self.reverseDecrypt == True):
                    eoi = 1
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[itype],finalSgtSize,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(itype,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
                self._pp("DAT = {0}".format(self._zpad(msg[int(fullSegment*self.param['maxSizeSegment']*2):],self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
        elif (opcode in [self.OPCODE_PRNG]):
            itype = self.STYPE_PLAINTEXT
            (eot, eoi) = (1, 1)
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[itype],sizeInput,pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(itype,pad,eoi,eot,sizeInput,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)

        ### Data (Output)
        if  (sizeOutput > 0) and \
            (opcode in [self.OPCODE_ENCRYPT, self.OPCODE_DECRYPT, \
                        self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, \
                        self.OPCODE_PRNG]):
            if opcode == self.OPCODE_DECRYPT or opcode == self.OPCODE_AEAD_DECRYPT:
                otype = self.STYPE_PLAINTEXT
            else:
                otype = self.STYPE_CIPHERTEXT

            eot             = 0
            fullSegment     = int(sizeOutput/self.param['maxSizeSegment'])
            sgtSize         = self.param['maxSizeSegment']
            finalSgtSize    = sizeOutput % self.param['maxSizeSegment']
            # Full blocks
            for i in range(0,fullSegment):
                if i == fullSegment-1 and finalSgtSize == 0:
                    (eot, eoi) = (1, 1)
                if opcode in [self.OPCODE_AEAD_ENCRYPT]:
                    eoi = 0
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[otype],sgtSize,pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(otype,pad,eoi,eot,sgtSize,self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(encrypted[int(i*self.param['maxSizeSegment']*2):int((i+1)*self.param['maxSizeSegment']*2)]), self.FILE_OUTPUT)

            # Partial blocks
            if finalSgtSize > 0:
                (eot, eoi) = (1, 1)
                if opcode in [self.OPCODE_AEAD_ENCRYPT]:
                    eoi = 0
                self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[otype],finalSgtSize,pad,eoi,eot), self.FILE_OUTPUT)
                self._pp("HDR = {0}".format(self.getHdrCmd(otype,pad,eoi,eot,finalSgtSize,self.param['sizePIO'])), self.FILE_OUTPUT)
                self._pp("DAT = {0}".format(self._xpad(encrypted[fullSegment*self.param['maxSizeSegment']*2:],self.param['sizePIO'])), self.FILE_OUTPUT)


        ### Tag
        if opcode in [self.OPCODE_AEAD_ENCRYPT, self.OPCODE_AEAD_DECRYPT, self.OPCODE_MAC, self.OPCODE_HASH]:
            if not ((opcode == self.OPCODE_AEAD_DECRYPT) and (self.reverseDecrypt == True) and (self.cipherTextMode == 2)):
                eoi = 1
                self._genTagSegment(tag, opcode, pad, eoi, eot)




    ### Generate Tag segment
    def _genTagSegment(self, tag, opcode, pad, eoi, eot):
        # Input
        if opcode == self.OPCODE_AEAD_DECRYPT:
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_TAG],int(len(tag)/2),pad,eoi,eot), self.FILE_PUBLIC_INPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_TAG,pad,eoi,eot,len(tag)/2,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
            self._pp("DAT = {0}".format(self._xpad(tag,self.param['sizePIO'])), self.FILE_PUBLIC_INPUT)
        # Output
        if opcode != self.OPCODE_AEAD_DECRYPT:  # No output TAG for decryption
            self._pp("# {0:03} : SgtHdr      (Size={2:4}) (PAD={3}) (EOI={4}) (EOT={5}) (SgtType={1})".format(self.msgid, self.CODE_STYPE[self.STYPE_TAG],int(len(tag)/2),pad,eoi,eot), self.FILE_OUTPUT)
            self._pp("HDR = {0}".format(self.getHdrCmd(self.STYPE_TAG,pad,eoi,eot,len(tag)/2,self.param['sizePIO'])), self.FILE_OUTPUT)
            self._pp("DAT = {0}".format(self._xpad(tag,self.param['sizePIO'])), self.FILE_OUTPUT)


    ### Main calling function
    def genTestVectors(self, sampleSize):
        """ Generate Test Vectors """
        # Print settings
        self._printInfo()
        random.seed()
        for i in range(0,sampleSize):
            sizeAD  = random.randrange(self.param['minSizeAD'], self.param['maxSizeAD']+1, 1)
            sizeMsg = random.randrange(self.param['minSizeM'],  self.param['maxSizeM']+1, 1)
            newKey  = random.getrandbits(1)
            if self.param['opcode'] != False:
                if len(self.param['opcode']) > 1:
                    opcode = self.param['opcode'][random.randrange(len(self.param['opcode']))]
                else:
                    opcode = self.param['opcode']
            else:
                opcode = random.getrandbits(1)+2 # Make sure that the opcode is always AEAD for now

            if (self.param['testMode'] != False):
                if (i % 2 == 1):
                    opcode = 3
                else:
                    opcode = 2
                if i in [0, 1]:
                    sizeMsg = 0
                    sizeAD  = 1                
                elif i in [2, 3]:
                    sizeMsg = 1
                    sizeAD  = 0
                elif i in [4, 5]:
                    sizeMsg = 1
                    sizeAD  = 1
                elif i in [6, 7]:
                    sizeMsg = self.blockSize
                    sizeAD  = sizeMsg
                elif i in [8, 9]:
                    sizeMsg = self.blockSize-1
                    sizeAD  = sizeMsg
                elif i in [10, 11]:
                    sizeMsg = 1+self.blockSize
                    sizeAD  = sizeMsg
                elif i in [12, 13]:
                    sizeMsg = 2*self.blockSize
                    sizeAD  = sizeMsg
                elif i in [14, 15]:
                    sizeMsg = random.randrange(0, 2*self.blockSize, 1)
                    sizeAD  = sizeMsg
                    while sizeAD == sizeMsg:
                        sizeAD = random.randrange(0, 2*self.blockSize, 1)
                else:
                    sizeMsg = ((i-12)/2+2)*self.blockSize
                    sizeAD  = sizeMsg

            print("Generating message {}/{} {} [AD Sizes={:< 2},Msg Size={}]".format(i+1,sampleSize,self.CODE_OPCODE[opcode],int(sizeAD),int(sizeMsg)))
            self.genTV(self.param['algorithm'], opcode, self.param['sizeNPUB'], int(sizeAD), int(sizeMsg), newKey)
        self._pp("###EOF", self.FILE_ALL)
        if self.param['stdout'] == False:
            self._fileIO('close')
        print("Terminating AETVgen.py")

    ### Alternative Calling function for mixed functions
    def genMixTestVectors(self, sampleSize):
        """ Generate mixed functions test vectors """
        self._printInfo(['opcode'])
        # Print settings
        self._printInfo()
        random.seed()
        for i in range(0,sampleSize):
            random.jumpahead(1)
            sizeAD = 0
            sizeNPUB = 0
            sizeP  = random.randrange(self.param['minSizeM'],  self.param['maxSizeM']+1, 1)
            newKey = random.getrandbits(1)
            algorithm = self.param['algorithm'][random.randrange(len(self.param['algorithm']))]
            if   algorithm.lower() in ['aes-gcm', 'aes_gcm', 'icepole', 'keccak-ae', 'joltik']:
                sizeNPUB = self.param['sizeNPUB']
                sizeAD = random.randrange(self.param['minSizeAD'], self.param['maxSizeAD']+1, 1)
                if random.getrandbits(1):
                    opcode = self.OPCODE_AEAD_ENCRYPT
                else:
                    opcode = self.OPCODE_AEAD_DECRYPT
            elif algorithm.lower() in ['aes-cmac', 'aes_cmac', 'keccak-mac']:
                opcode = self.OPCODE_MAC
            elif algorithm.lower() in ['aes-ecb', 'aes_ecb', 'aes-ctr']:
                # AES-CTR requires NPUB
                if algorithm.lower() in ['aes-ctr']:
                    sizeNPUB = self.param['sizeNPUB']
                if random.getrandbits(1):
                    opcode = self.OPCODE_ENCRYPT
                else:
                    opcode = self.OPCODE_DECRYPT
            elif algorithm.lower() in ['aes-hash', 'aes_hash', 'keccak-hash', 'sha3']:
                opcode = self.OPCODE_HASH
            elif algorithm.lower() in ['aes-fortuna', 'keccak-prng']:
                sizeNPUB = self.param['sizeNPUB']
                opcode = self.OPCODE_PRNG
                if algorithm.lower() in ['keccak-prng']:
                    newKey = 0
            else:
                raise ValueError('Error! Unsupported algorithm for genMixTestVectors')
            print("Generating message {0}/{1} with size AD={2}, size Msg={3}".format(i+1,sampleSize,sizeAD,sizeP))
            self.genTV(algorithm, opcode, sizeNPUB, sizeAD, sizeP, newKey)
        self._pp("###EOF", self.FILE_ALL)
        if self.param['stdout'] == False:
            self._fileIO('close')
        print("Terminating AETVgen.py")

    ### =================
    ### Helper functions
    ### =================
    def _xpad(self, text, iowidth):
        """ 'X' pad the text to match the I/O width """
        return text + 'X'*((-len(text)) % (iowidth*2))

    def _zpad(self, text, iowidth):
        """ Zero pad the text to match the I/O width """
        return text + '0'*((-len(text)) % (iowidth*2))

    def _calcIOFill(self, filled, iowidth):
        return (iowidth*2-filled) % (iowidth*2)

    def _fileIO(self, op='open'):
        if op=='open':
            print ("Opening files...")
            self.pdi = open(self.param['filePDI'], "wt")
            self.sdi = open(self.param['fileSDI'], "wt")
            self.do  = open(self.param['fileDO'],  "wt")
        else:
            print ("Closing files...")
            self.pdi.close()
            self.sdi.close()
            self.do.close()

    def _genRandomData(self, size, base='HEX'):
        """ Generate random data """
        if (size == 0):
            return ("00")
        else:
            if base=='HEX':
                return ("%0*X" % (size*2, random.randrange(256**size)))
            else:
                print ("Error! Unsupported <base> for function _genRandomData(<size>, <base>)")
                print ("base = %s" % (base))

    def _printCallInfo(self, algorithm, opcode, key, npub, ad, msg, encrypted, tag):
        print ("\n\n")
        print ("{0} - {1}".format(algorithm, self.CODE_OPCODE[opcode]))
        print ("Key\t\t:{0}".format(key))
        print ("NPUB\t\t:{0}".format(npub))
        print ("Ad\t\t:{0}".format(ad))
        print ("Msg\t\t:{0}".format(msg))
        print ("Enc(Msg)\t:{0}".format(encrypted))
        print ("Tag\t\t:{0}".format(tag))
        print ("Len(Msg)\t:{0}".format(len(msg)/2))
        print ("Len(encrypted)\t:{0}".format(len(encrypted)/2))
        print ("Len(tag)\t:{0}".format(len(tag)/2))

    def _printInfo(self, additionalIgnores=[]):
        ignoreParams = ['stdout', 'filePDI', 'fileSDI', 'fileDO'] + additionalIgnores
        if self.param['stdout'] == False:
            self._fileIO('open')

        for target in self.FILE_SECRET, self.FILE_PUBLIC_INPUT, self.FILE_OUTPUT:
            self._pp("# ===================================================", target)
            if self.param['stdout'] == False:
                if target == self.FILE_SECRET:
                    self._pp("# {0}".format(self.sdi.name),                   target)
                elif target == self.FILE_PUBLIC_INPUT:
                    self._pp("# {0}".format(self.pdi.name),                   target)
                elif target == self.FILE_OUTPUT:
                    self._pp("# {0}".format(self.do.name),                    target)
            self._pp("# This file was generated using AETVgen.py script",     target)
            self._pp("# with the following parameters:",                      target)
            self._pp("#",                                                     target)
            for (key, value) in sorted(self.param.items(), key=lambda key_value: key_value[0]):
                if key not in ignoreParams:
                    self._pp("# {0:15} {1}".format(key, value), target)
            self._pp("#",                                                     target)
            self._pp("# ===================================================", target)

    def _pp(self, text, target=FILE_ALL):
        """ Pretty printing """
        if self.param['stdout'] == True:
            if target == self.FILE_ALL:
                tgt = "pdi, sdi, do"
            elif target == self.FILE_INPUT:
                tgt = "pdi, sdi"
            elif target == self.FILE_OUTPUT:
                tgt = "do"
            elif target == self.FILE_PUBLIC:
                tgt = "pdi, do"
            elif target == self.FILE_PUBLIC_INPUT:
                tgt = "pdi"
            elif target == self.FILE_SECRET:
                tgt = "sdi"
            else:
                tgt = ""
            print ("{0:12} | {1}".format(tgt, text) )
        elif target == self.FILE_ALL:
            self.pdi.write(text + "\n")
            self.sdi.write(text + "\n")
            self.do.write(text + "\n")
        elif target == self.FILE_INPUT:
            self.pdi.write(text + "\n")
            self.sdi.write(text + "\n")
        elif target == self.FILE_OUTPUT:
            self.do.write(text + "\n")
        elif target == self.FILE_PUBLIC:
            self.pdi.write(text + "\n")
            self.do.write(text + "\n")
        elif target == self.FILE_PUBLIC_INPUT:
            self.pdi.write(text + "\n")
        elif target == self.FILE_SECRET:
            self.sdi.write(text + "\n")
        else:
            raise ValueError("Invalid write target")

    def _debugGenTest(self):
        self._printInfo()
        self.param['sizeKey']        = 16
        self.param['sizeNPUB']         = 12
        self.param['sizeTag']        = 16
        sizeAD = 16
        sizeP  = 16
        newKey = 1
        self.genTV(self.param['algorithm'], 2, sizeAD, sizeP, newKey)

if __name__ == '__main__':
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSizeM': 1, 'maxSizeM': 64,
            'minSizeAD': 1, 'maxSizeAD': 64,
            'maxSizeSegment' : 32}
    aetv = AETVgen('aes-gcm', **args)
    aetv._debugGenTest()
