-------------------------------------------------------------------------------
--! @file       PostProcessor_Control.vhd
--! @brief      Control unit for post-processor
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2015 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @version    1.0b1
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity PostProcessor_Control is
    generic (
        G_W                     : integer := 64;      --! Output width (bits)
        G_DBLK_SIZE             : integer := 128;     --! Block size (bits)
        G_BS_BYTES              : integer := 4;       --! The number of bits required to hold block size expressed in bytes = log2_ceil(block_size/8)
        G_TAG_SIZE              : integer := 128;     --! Tag size (bits)
        G_REVERSE_DBLK          : integer := 0;       --! Reverse order of message block
        G_CIPHERTEXT_MODE       : integer := 0        --! Ciphertext processing mode
    );
    port (
        --! =================
        --! Global Signals
        --! =================
        --! Global signals
        clk                 : in  std_logic;
        rst                 : in  std_logic;

        --! =================
        --! External Signals
        --! =================
        do_ready            : in  std_logic;                                                           --! Output FIFO ready
        do_valid            : out std_logic;
        bypass_fifo_data    : in  std_logic_vector(G_W                                -1 downto 0);    --! Bypass FIFO data
        bypass_fifo_empty   : in  std_logic;                                                           --! Bypass FIFO empty
        bypass_fifo_rd      : out std_logic;                                                           --! Bypass FIFO read
        bdo_ready           : out std_logic;                                                           --! Output BDO ready (Let crypto core knows that it's ready to accept data)
        bdo_write           : in  std_logic;                                                           --! Write to output BDO
        tag_ready           : out std_logic;                                                           --! Output TAG ready (Let crypto core knows that it's ready to accept data)
        tag_write           : in  std_logic;                                                           --! Write to output TAG
        msg_auth_done       : in  std_logic;
        msg_auth_valid      : in  std_logic;

        --! =================
        --! Controls
        --! =================
        bdo_shf             : out std_logic;                                                           --! Shift output BDO
        tag_shf             : out std_logic;                                                           --! Shift tag register

        is_lsm              : out std_logic;                                                           --! Signal indicating that it's the last segment of message
        sel_instr           : out std_logic;                                                           --! Increment instruction Opcode by 1
        sel_hdr             : out std_logic;                                                           --! Header type selection
        sel_do              : out std_logic_vector(  2                                -1 downto 0);    --! Output selection
        sel_sw              : out std_logic_vector(  2                                -1 downto 0);    --! Segment type encoding selection
        sel_error           : out std_logic;                                                           --! Error output selection (when a tag is not matched).
        is_i_type           : out std_logic_vector(  4                                -1 downto 0);    --! Segment type identifier 1 (Cipher/Message)
        is_i_nsec           : out std_logic_vector(  4                                -1 downto 0);    --! Segment type identifier 2 (Secret number message / Enecrypted secret number message)
        msg_id              : out std_logic_vector(  8                                -1 downto 0);    --! Message ID

        en_size             : out std_logic;
        clr_size            : out std_logic;
        sel_do2             : out std_logic;

        en_partial_word     : out std_logic;
        data_bytes          : out std_logic_vector(log2_ceil(G_W/8)                   -1 downto 0);

        aux_fifo_ctrl       : out std_logic_vector(3 downto 0);
        aux_fifo_status     : in  std_logic_vector(2 downto 0)
    );
end PostProcessor_Control;

architecture behavior of PostProcessor_Control is
    --! Function and constants declaration
    constant LOG2_WD8           : integer                                           := log2_ceil(G_W/8);
    constant ONES               : std_logic_vector(G_DBLK_SIZE-1 downto 0)          := (others => '1');
    constant ZEROS              : std_logic_vector(G_DBLK_SIZE-1 downto 0)          := (others => '0');
    constant CNTR_WIDTH         : integer := get_cntr_width(G_W);

    type state_type     is (S_WAIT_INSTR, S_READ_INSTR, S_WAIT_HDR, S_READ_HDR,
                            S_WAIT_BYPASS, S_WAIT_BDO, S_GEN_TAG_HDR, S_WRITE_TAG,
                            S_WAIT_MSG_AUTH, S_WRITE_TAG_ERROR, S_ERROR);

    signal state                : state_type;
    signal nstate               : state_type;

    signal fifo_write_pre       : std_logic;
    signal fifo_write_now       : std_logic;
    signal fifo_write_r         : std_logic;
    signal sel_do_pre           : std_logic_vector(  2                                -1 downto 0);
    signal sel_error_pre        : std_logic;

    signal tag_count            : std_logic_vector(log2_ceil(G_TAG_SIZE/8)            -1 downto 0);
    signal tag_empty            : std_logic;
    signal tag_shf_pre          : std_logic;

    signal bdo_count            : std_logic_vector(G_BS_BYTES                         -1 downto 0);
    signal bdo_empty            : std_logic;
    signal bdo_shf_pre          : std_logic;
    signal bdo_clr_pre          : std_logic;

    signal opcode               : std_logic_vector(  4                                -1 downto 0);
    signal sgmt_stype           : std_logic_vector(  4                                -1 downto 0);
    signal sgmt_lsm             : std_logic;
    signal sgmt_lst             : std_logic;

    signal en_instr_flag        : std_logic;
    signal is_decrypt           : std_logic;
    signal is_ae                : std_logic;
    signal en_sgmt_status        : std_logic;
    signal clr_hdr_status       : std_logic;

    signal instr_decrypt        : std_logic;
    signal instr_ae             : std_logic;
    signal sgmt_msg_type        : std_logic_vector(  4                                -1 downto 0);
    signal sgmt_nsec_type       : std_logic_vector(  4                                -1 downto 0);
    signal sgmt_nsec_flag       : std_logic;
    signal msg_end              : std_logic;
    signal msg_id_r             : std_logic_vector(  8                                -1 downto 0);

    signal counter_load_tag     : std_logic;
    signal counter_load         : std_logic;
    signal counter_en           : std_logic;
    signal counter              : std_logic_vector(CNTR_WIDTH                         -1 downto 0);

    signal sw_stype             : std_logic;
    signal sw_nsec              : std_logic;

    signal check_tag            : std_logic;
    signal clr_check_tag        : std_logic;
    signal restore_state        : std_logic;
    signal do_write_pre         : std_logic;

    signal clr_status           : std_logic;
    signal set_sgmt_tag_flag    : std_logic;
    signal sgmt_tag_flag        : std_logic;

    signal set_no_write         : std_logic;
    signal no_write             : std_logic;

    signal msg_auth_done_r      : std_logic;
    signal msg_auth_valid_r     : std_logic;


    signal fifo_save_state      : std_logic;
    signal fifo_restore_state   : std_logic;
    signal fifo_write           : std_logic;
    signal fifo_read            : std_logic;
    signal fifo_unread_avail    : std_logic;
    signal fifo_empty           : std_logic;
    signal fifo_full            : std_logic;
begin
    fifo_unread_avail <= aux_fifo_status(0);
    fifo_empty        <= aux_fifo_status(1);
    fifo_full         <= aux_fifo_status(2);
    aux_fifo_ctrl     <= fifo_read & fifo_write & fifo_restore_state & fifo_save_state;

    --! Output
    is_lsm    <= '1' when (state = S_READ_HDR and sgmt_lsm = '1' and not ((sgmt_stype = ST_MESSAGE and instr_ae = '1') or (sgmt_stype = ST_AD))) or
                          (state = S_READ_HDR and sgmt_lst = '1' and sgmt_stype = ST_CIPHER) or   --! Ciphertext input.
                          (state = S_GEN_TAG_HDR) else '0';
    sel_sw    <= sw_stype & sw_nsec;
    is_i_type <= sgmt_msg_type;
    is_i_nsec <= sgmt_nsec_type;
    msg_id    <= msg_id_r;

    fifo_save_state     <= is_decrypt;
    fifo_restore_state  <= restore_state;

    --! Format decoder
    opcode    <= bypass_fifo_data(G_W-12-1 downto G_W-16);
    sgmt_stype <= bypass_fifo_data(G_W- 8-1 downto G_W-12);
    sgmt_lst   <= bypass_fifo_data(G_W-15-1);
    sgmt_lsm   <= bypass_fifo_data(G_W-14-1);

    data_bytes <= counter(log2_ceil(G_W/8)-1 downto 0);

    --! Registers
    procRegs:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                state           <= S_WAIT_INSTR;
                instr_ae        <= '0';
                sgmt_msg_type    <= (others => '0');
                msg_end         <= '0';
                counter         <= (others => '0');
                msg_id_r        <= (others => '0');
                bdo_empty       <= '1';
                tag_empty       <= '1';
                check_tag     <= '0';
                msg_auth_done_r <= '0';
                tag_count       <= (others => '0');
                bdo_count       <= (others => '0');
                no_write        <= '0';
            else
                state <= nstate;
                if en_instr_flag = '1' then
                    instr_ae   <= is_ae;
                    if is_decrypt = '1' then
                        sgmt_msg_type     <= ST_MESSAGE;
                        sgmt_nsec_type    <= ST_NSEC;
                        instr_decrypt    <= '1';
                    else
                        sgmt_msg_type     <= ST_CIPHER;
                        sgmt_nsec_type    <= ST_NSEC_CIPH;
                        instr_decrypt    <= '0';
                    end if;
                end if;

                if (en_instr_flag = '1') then
                    if (is_decrypt = '1') then
                        check_tag <= '1';
                    else
                        check_tag <= '0';
                    end if;
                elsif (clr_check_tag = '1') then
                    check_tag <= '0';
                end if;

                if clr_hdr_status = '1' then
                    msg_end   <= '0';
                elsif en_sgmt_status = '1' then
                    msg_id_r  <= bypass_fifo_data(G_W-1 downto G_W-LEN_MSG_ID);
                    msg_end   <= sgmt_lsm;
                end if;

                if counter_load_tag = '1' then
                    counter <= std_logic_vector(to_unsigned(G_TAG_SIZE/8, G_W-16));
                elsif counter_load = '1' then
                    if (G_CIPHERTEXT_MODE = 2) then
                        if (sgmt_stype = ST_MESSAGE) then
                            counter <= (bypass_fifo_data(CNTR_WIDTH-1 downto G_BS_BYTES) + 1) & ZEROS(G_BS_BYTES-1 downto 0);
                        else
                            counter <= bypass_fifo_data(CNTR_WIDTH-1 downto 0);
                        end if;
                    else
                        counter <= bypass_fifo_data(CNTR_WIDTH-1 downto 0);
                    end if;
                elsif counter_en = '1' then
                    counter <= counter - G_W/8;
                end if;

                if state = S_READ_HDR then
                    if sw_nsec = '1' then
                        sgmt_nsec_flag <= '1';
                    else
                        sgmt_nsec_flag <= '0';
                    end if;
                end if;

                --! Keeps track whether the BDO register is empty
                if (bdo_write = '1') then
                    bdo_empty      <= '0';
                elsif ((bdo_shf_pre = '1' and counter = 1) or (bdo_count = 0 and bdo_shf_pre = '1') or (bdo_clr_pre = '1')) then
                    bdo_empty      <= '1';
                end if;

                --! Keeps track of available bytes in BDO register
                if (bdo_write = '1') then
                    bdo_count      <= std_logic_vector(to_unsigned(G_DBLK_SIZE/G_W-1, G_BS_BYTES));
                elsif (bdo_shf_pre = '1') then
                    bdo_count      <= bdo_count - 1;
                end if;

                --! Keeps track whether the TAG register is empty
                if (tag_write = '1') then
                    tag_empty      <= '0';
                elsif ((tag_shf_pre = '1' and counter = 1) or (tag_count = 0 and tag_shf_pre = '1')) then
                    tag_empty      <= '1';
                end if;

                --! Keeps track of available bytes in TAG register
                if (tag_write = '1') then
                    tag_count      <= std_logic_vector(to_unsigned(G_TAG_SIZE/G_W-1, log2_ceil(G_TAG_SIZE/8)));
                elsif (tag_shf_pre = '1') then
                    tag_count      <= tag_count - 1;
                end if;


                if (set_sgmt_tag_flag = '1') then
                    sgmt_tag_flag <= '1';
                elsif (clr_status = '1') then
                    sgmt_tag_flag <= '0';
                end if;

                if (set_no_write = '1') then
                    no_write <= '1';
                elsif (clr_status = '1') then
                    no_write <= '0';
                end if;

                --! Capture the tag comparison flag and result
                if (clr_check_tag = '1') then
                    msg_auth_done_r   <= '0';
                elsif (msg_auth_done = '1') then
                    msg_auth_done_r   <= '1';
                    msg_auth_valid_r  <= msg_auth_valid;
                end if;

                tag_shf       <= tag_shf_pre;
                bdo_shf       <= bdo_shf_pre;
                sel_do        <= sel_do_pre;
                fifo_write_r  <= fifo_write_pre;
                sel_error     <= sel_error_pre;
            end if;
        end if;
    end process;
    fifo_write <= fifo_write_r or fifo_write_now;

    --! Controller
    procFSM:
    process( state, bypass_fifo_empty, bypass_fifo_data, bdo_empty, fifo_full, sgmt_nsec_flag,
             counter, opcode, sgmt_stype, msg_end, instr_ae, check_tag, sgmt_tag_flag,
             tag_empty, msg_auth_valid_r, msg_auth_done_r, instr_decrypt)
    begin
        sel_do_pre          <= "00";
        fifo_write_pre      <= '0';
        fifo_write_now      <= '0';
        bypass_fifo_rd      <= '0';
        is_ae               <= '0';
        is_decrypt          <= '0';
        en_instr_flag       <= '0';
        en_sgmt_status       <= '0';
        clr_hdr_status      <= '0';
        counter_load_tag    <= '0';
        counter_load        <= '0';
        counter_en          <= '0';
        sw_stype            <= '0';
        sel_instr           <= '0';
        sel_hdr             <= '0';
        sel_error_pre       <= '0';
        sw_nsec             <= '0';
        bdo_shf_pre         <= '0';
        bdo_clr_pre         <= '0';
        restore_state       <= '0';
        clr_check_tag       <= '0';
        tag_shf_pre         <= '0';
        set_sgmt_tag_flag      <= '0';
        clr_status          <= '0';
        en_partial_word     <= '0';
        set_no_write        <= '0';
        nstate              <= state;

        case state is
            when S_WAIT_INSTR =>
                if (bypass_fifo_empty = '0' and fifo_full = '0') then
                    nstate          <= S_READ_INSTR;
                    bypass_fifo_rd  <= '1';
                    clr_hdr_status      <= '1';
                    bdo_clr_pre     <= '1';
                    clr_status      <= '1';
                end if;

            when S_READ_INSTR =>
                if (opcode = OP_ACT_KEY) then
                    nstate          <= S_WAIT_INSTR;
                else
                    nstate          <= S_WAIT_HDR;
                    en_instr_flag   <= '1';
                end if;
                if (opcode = OP_AE_DEC or opcode = OP_DEC) then
                    is_decrypt      <= '1';
                end if;
                if (opcode = OP_AE_ENC or opcode = OP_AE_DEC) then
                    is_ae           <= '1';
                end if;
                if (opcode = OP_AE_ENC or opcode = OP_ENC) then
                    fifo_write_now <= '1';
                    sel_instr      <= '1';
                end if;

            when S_WAIT_HDR =>
                if (bypass_fifo_empty = '0' and fifo_full = '0') then
                    nstate          <= S_READ_HDR;
                    bypass_fifo_rd  <= '1';
                end if;

            when S_READ_HDR =>
                if ((sgmt_stype = ST_TAG or sgmt_stype = ST_NPUB) and instr_decrypt = '1') then
                    fifo_write_now      <= '0';
                else
                    fifo_write_now      <= '1';
                end if;
                counter_load        <= '1';
                en_sgmt_status      <= '1';
                sel_hdr             <= '1';

                if (sgmt_stype = ST_MESSAGE or sgmt_stype = ST_CIPHER) then
                    sw_stype        <= '1';
                    nstate          <= S_WAIT_BDO;
                elsif (sgmt_stype = ST_NSEC or sgmt_stype = ST_NSEC_CIPH) then
                    sw_stype        <= '1';
                    sw_nsec         <= '1';
                    nstate          <= S_WAIT_BDO;
                elsif (sgmt_stype = ST_NPUB) then
                    nstate          <= S_WAIT_BYPASS;
                    if (instr_decrypt = '1') then
                        set_no_write    <= '1';
                    end if;                    
                elsif (sgmt_stype = ST_AD) then
                    nstate          <= S_WAIT_BYPASS;                    
                elsif (sgmt_stype = ST_TAG) then
                    set_sgmt_tag_flag  <= '1';
                    nstate          <= S_WAIT_BYPASS;
                else
                    nstate          <= S_ERROR;
                end if;

            when S_WAIT_BYPASS =>
                if (bypass_fifo_empty = '0' and fifo_full = '0') then                    
                    counter_en      <= '1';
                    bypass_fifo_rd  <= '1';
                    if (sgmt_tag_flag = '0' and  --! No output write if, sgmt = tag
                        no_write = '0')          --! No output write if, sgmt = IV and decrypt = 1 
                    then        
                        fifo_write_pre      <= '1';
                    end if;

                    if counter <= G_W/8 then
                        clr_status <= '1';
                        if msg_end = '1' then                            
                            if  (sgmt_tag_flag = '0') and --! Special case: No message / no AD
                                (instr_decrypt = '0')     --! Special case: No message
                            then
                                nstate  <= S_GEN_TAG_HDR;       
                            else
                                nstate  <= S_WAIT_MSG_AUTH;
                            end if;
                        else
                            nstate  <= S_WAIT_HDR;
                        end if;
                    end if;
                end if;

            when S_WAIT_BDO =>
                if (bdo_empty = '0' and fifo_full = '0') then
                    bdo_shf_pre     <= '1';
                    fifo_write_pre  <= '1';
                    sel_do_pre      <= "01";
                    counter_en      <= '1';
                    if (counter < G_W/8) then
                        en_partial_word <= '1';
                    end if;
                    if (counter <= G_W/8) then
                        if (msg_end = '1') then
                            bdo_clr_pre <= '1';
                            if (instr_ae = '1' and sgmt_nsec_flag = '0') then
                                if (G_CIPHERTEXT_MODE = 1 and G_REVERSE_DBLK = 1 and instr_decrypt = '1') then
                                    nstate  <= S_WAIT_MSG_AUTH;
                                else
                                    nstate  <= S_GEN_TAG_HDR;
                                end if;
                            elsif sgmt_nsec_flag = '1' then
                                nstate  <= S_WAIT_HDR;
                            else
                                nstate  <= S_WAIT_INSTR;
                            end if;
                        else
                            nstate      <= S_WAIT_HDR;
                        end if;
                    end if;
                end if;

            when S_GEN_TAG_HDR  =>
                nstate              <= S_WRITE_TAG;
                fifo_write_pre      <= '1';
                sel_do_pre          <= "10";
                counter_load_tag    <= '1';

            when S_WRITE_TAG =>
                if (tag_empty = '0' and fifo_full = '0') then
                    tag_shf_pre    <= '1';
                    sel_do_pre      <= "11";
                    if (counter /= 0) then
                        counter_en      <= '1';
                    end if;
                    fifo_write_pre  <= '1';
                    if (counter <= G_W/8) then
                        nstate          <= S_WAIT_INSTR;
                    end if;
                end if;

            when S_WAIT_MSG_AUTH =>
                if (msg_auth_done_r = '1') then
                    clr_check_tag       <= '1';
                    if (msg_auth_valid_r = '1') then
                        nstate          <= S_WAIT_INSTR;
                    else
                        restore_state   <= '1';
                        nstate          <= S_WRITE_TAG_ERROR;
                    end if;
                end if;

            when S_WRITE_TAG_ERROR =>
                sel_do_pre      <= "10";
                sel_error_pre   <= '1';
                fifo_write_pre  <= '1';
                nstate          <= S_WAIT_INSTR;

            when S_ERROR =>

        end case;
    end process;

    bdo_ready <= bdo_empty;
    tag_ready <= tag_empty;


    G_CIPH01: if G_CIPHERTEXT_MODE /= 2 generate
        procFIFOCtrl:
        process(clk)
        begin
            if rising_edge(clk) then
                if (fifo_empty = '0' and do_ready = '1' and (check_tag = '0' or fifo_unread_avail = '1')) then
                    fifo_read       <= '1';
                    do_write_pre    <= '1';
                else
                    fifo_read       <= '0';
                    do_write_pre    <= '0';
                end if;
                do_valid <= do_write_pre;
            end if;
        end process;
    end generate;

    G_CEXP2 : if G_CIPHERTEXT_MODE = 2 generate
        procFIFOCtrl:
        process(clk)
        begin
            if rising_edge(clk) then
                if (fifo_empty = '0' and do_ready = '1' and (check_tag = '0' or fifo_unread_avail = '1')) then
                    fifo_read       <= '1';
                    do_write_pre    <= '1';
                else
                    fifo_read       <= '0';
                    do_write_pre    <= '0';
                end if;
                do_valid <= do_write_pre;
            end if;
        end process;

        -- if (bdo_write = '1')
    end generate;


end behavior;
