-------------------------------------------------------------------------------
--! @file       PreProcessor_Control.vhd
--! @brief      Control unit for the pre-processor
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2015 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @version    1.0b1
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity PreProcessor_Control is
    generic (
        G_W                      : integer := 64;   --! Public data width (bits)
        G_SW                     : integer := 64;   --! Secret data width (bits)
        G_NPUB_SIZE              : integer := 128;  --! NPUB width (bits)
        G_ABLK_SIZE              : integer := 128;  --! Authenticated Data Block size (bits)
        G_DBLK_SIZE              : integer := 128;  --! Data Block size (bits)
        G_TAG_SIZE               : integer := 128;  --! Tag size (bits)
        G_BS_BYTES               : integer := 4;    --! The number of bits required to hold block size expressed in bytes = log2_ceil(block_size/8)
        G_KEY_SIZE               : integer := 128;  --! Key size (bits)
        G_PAD                    : integer := 0;    --! Enable padding
        G_PAD_AD                 : integer := 1;    --! (G_PAD's sub option) Enable AD Padding
        G_PAD_D                  : integer := 1;    --! (G_PAD's sub option) Enable Data padding
        G_CIPHERTEXT_MODE        : integer := 0;
        G_PLAINTEXT_MODE         : integer := 0;
        G_REVERSE_DBLK           : integer := 0;    --! Reverse block order (for message only)
        G_DISABLE_NPUB_PORT      : integer := 0;    --! Disable NPUB related port(s)
        G_KEYAK                  : integer := 0     --! Special Keyak only mode
    );
    port (
        --! =================
        --! External Signals
        --! =================
        --! Global signals
        clk                 : in  std_logic;
        rst                 : in  std_logic;

        --! Public signals
        pdi                 : in  std_logic_vector(G_W                        -1 downto 0);
        pdi_valid           : in  std_logic;
        pdi_ready            : out std_logic;

        --! Secret signals
        sdi                 : in  std_logic_vector(G_SW                       -1 downto 0);
        sdi_valid           : in  std_logic;
        sdi_ready            : out std_logic;

        error               : out std_logic;

        --! =================
        --! Crypto Core Signals
        --! =================
        --! control signals
        key_ready           : out std_logic;                                                      --! Indicates that the key is ready
        key_needs_update    : out std_logic;                                                      --! Indicates that the key needs update and should be acknowledge by the core via key_updated signal
        key_updated         : in  std_logic;                                                      --! Key has been updated
        npub_ready            : out std_logic;                                                      --! NPUB ready signal

        bdi_ready           : out std_logic;                                                      --! Block ready
        bdi_proc            : out std_logic;                                                      --! Block processing
        bdi_ad              : out std_logic;                                                      --! Input block is an authenticated data
        bdi_nsec            : out std_logic;                                                      --! Input block is a secret message number
        bdi_decrypt         : out std_logic;                                                      --! Decryption operation
        bdi_pad             : out std_logic;                                                      --! Last block of segment type contain padding
        bdi_eot             : out std_logic;                                                      --! Last block of segment type (end-of-type)
        bdi_eoi             : out std_logic;                                                      --! Last block of message (end-of-input)
        bdi_nodata          : out std_logic;                                                      --! Control signal indicating that there's no plain-text or authenticated data. The unit should generate a tag right away.
        bdi_read            : in  std_logic;                                                      --! Handshake signal indicating that the data block has been read
        bdi_size            : out std_logic_vector(G_BS_BYTES                 -1 downto 0);       --! Block size signal. Note: 0 = Full block.
        exp_tag_ready       : out std_logic;                                                      --! Expected tag is ready
        msg_auth_done       : in  std_logic;                                                      --! Tag comparison completion handshake

        bypass_fifo_full    : in  std_logic;                                                      --! An input signal indicating that the bypass FIFO is full
        bypass_fifo_wr      : out std_logic;                                                      --! An output signal for writing data to bypass FIFO

        --! =================
        --! Internal Signals
        --! =================
        pad_shift           : out std_logic_vector(log2_ceil(G_W/8)           -1 downto 0);
        en_data             : out std_logic;                                                      --! Shift data SIPO
        en_npub             : out std_logic;                                                      --! Shift NPUB SIPO
        en_key              : out std_logic;                                                      --! Shift Key SIPO
        sel_blank_pdi       : out std_logic;                                                      --! Select input data as blank (for filling in the remaining data within a block)
        clr_len             : out std_logic;                                                      --! Clear stored length (len_a and len_d)
        en_len_a            : out std_logic;                                                      --! Add authenticated data counter
        en_len_d            : out std_logic;                                                      --! Add data counter
        en_exp_tag          : out std_logic;                                                      --! Shift TAG SIPO
        size_dword          : out std_logic_vector(log2_ceil(G_W/8)              downto 0);      --! Size of data word
        en_last_word        : out std_logic;                                                      --! Last word of a block
        --! Pad related control
        pad_eot             : out std_logic;                                                      --! Padding is EOT
        pad_type_ad         : out std_logic;                                                      --! Padding is AD
        pad_enable          : out std_logic;                                                      --! Enable padding signal (indicates that the current word requires padding)
        en_pad_loc          : out std_logic;                                                      --! Save the padding location into a register
        --! Supplemental control
        key_updated_int     : out std_logic;                                                      --! Only used for Keyak
        sel_input           : out std_logic_vector(2 downto 0)                                    --! Sel input (used when (G_DBLK_SIZE mod G_W) > 0)
    );
end PreProcessor_Control;

architecture behavior of PreProcessor_Control is
    function max(a: integer; b: integer) return integer is
    begin
        if a > b then
            return a;
        else
            return b;
        end if;
    end function max;

    function getPartialConstant return integer is
    begin
        if (G_DBLK_SIZE MOD G_W) /= 0 then
            return 1;
        else
            return 0;
        end if;
    end function getPartialConstant;

    --! Constants declaration
    constant PARTIAL_LOAD       : integer := getPartialConstant;
    constant LOG2_W             : integer := log2_ceil(G_W/8);                                    --! LOG_2(G_W)
    constant LOG2_KEYWORDS      : integer := log2_ceil(G_KEY_SIZE/G_SW);                          --! Expected key words
    constant REGIV_WIDTH        : integer := get_width(G_NPUB_SIZE, G_W);                           --! Calculate the width of NPUB register
    constant CNTR_WIDTH         : integer := get_cntr_width(G_W);                                 --! Calculate the length of p_size register
    constant CNT_NPUB_WORDS     : integer := get_words(G_NPUB_SIZE, G_W);                           --! Calculate the number of words required for NPUB
    constant CNT_AD_WORDS       : integer := (G_ABLK_SIZE+(G_W-1))/G_W;                           --! Calculate the number of words required for data (rounded up)
    constant CNT_DATA_WORDS     : integer := (G_DBLK_SIZE+(G_W-1))/G_W;                           --! Calculate the number of words required for data (rounded up)
    constant CNT_TAG_WORDS      : integer := (G_TAG_SIZE+(G_W-1))/G_W;                            --! Calculate the number of words required for tag  (rounded up)
    constant BSHIFT_INPUT       : std_logic_vector(G_W/8              -1 downto 0) := std_logic_vector(to_unsigned(1,G_W/8));
    constant WCOUNT_SIZE	    : integer := log2_ceil(max(max(CNT_NPUB_WORDS, CNT_DATA_WORDS), CNT_TAG_WORDS)+1);
    constant ZEROS              : std_logic_vector(G_DBLK_SIZE-1 downto 0) := (OTHERS => '0');

    --! Types declaration
    type pstate_type is (SP_WAIT_INSTR,     SP_READ_INSTR,
                         SP_WAIT_HDR,       SP_READ_HDR,       SP_WAIT_NPUB,
                         SP_INIT_KEYAK0,    SP_INIT_KEYAK1,    SP_INIT_KEYAK2,
                         SP_WAIT_DATA,      SP_WAIT_DATA_READ, SP_WAIT_MSG_AUTH,
                         SP_ERROR);

    type sstate_type is (SS_WAIT_INSTR,     SS_READ_INSTR,
                         SS_WAIT_HDR,       SS_READ_HDR,
                         SS_WAIT_DATA);

    --! State signals
    signal pstate               : pstate_type;
    signal sstate               : sstate_type;

    signal set_npub_ready       : std_logic;
    signal clr_npub_ready       : std_logic;

    --! Key related signals and registers
    signal set_key_needs_update : std_logic;
    signal set_key_ready        : std_logic;
    signal key_ready_r	        : std_logic;
    signal key_needs_update_r   : std_logic;

    signal s_opcode             : std_logic_vector(LEN_OPCODE                 -1 downto 0);       --! Secret data opcode
    signal s_stype              : std_logic_vector(LEN_SMT_HDR                -1 downto 0);       --! Secret data segment type
    signal s_reg_stype          : std_logic_vector(LEN_SMT_HDR                -1 downto 0);       --! Secret data register for segment type
    signal s_size               : std_logic_vector(G_SW-16                    -1 downto 0);       --! Secret data segment size
    signal swrem                : std_logic_vector(LOG2_KEYWORDS              -1 downto 0);       --! Secret data words remaining

    --! Public data signals and registers
    signal p_reg_stype          : std_logic_vector(LEN_SMT_HDR                -1 downto 0);       --! Segment type register
    signal pwrem                : std_logic_vector(CNTR_WIDTH                 -1 downto 0);       --! Private data words remaining
    signal wcount               : std_logic_vector(WCOUNT_SIZE                -1 downto 0);       --! Word count

    signal p_opcode             : std_logic_vector(LEN_OPCODE                 -1 downto 0);       --! Opcode
    signal p_stype              : std_logic_vector(LEN_SMT_HDR                -1 downto 0);       --! Segment type
    signal p_lst                : std_logic;                                                      --! Last segment of the specified type flag
    signal p_lsm                : std_logic;                                                      --! Last segment of the message flag
    signal p_pad                : std_logic;                                                      --! A flag indicating of padding unit
    signal p_id                 : std_logic_vector(LEN_MSG_ID                 -1 downto 0);       --! ID
    signal p_size               : std_logic_vector(CNTR_WIDTH                 -1 downto 0);       --! Public data segment size
    signal p_size_r             : std_logic_vector(CNTR_WIDTH                 -1 downto 0);       --! Public data segment size register

    signal remainder            : std_logic_vector(CNTR_WIDTH                 -1 downto 0);
    --!     Data block status register
    signal is_ae_r              : std_logic;                                                      --! Current block is an authenticated encryption block

    --!     Data block status registers for external modules
    signal is_decrypt_r         : std_logic;                                                      --! Current message is a decryption
    signal is_ad_r              : std_logic;                                                      --! Current block is an authenticated data block
    signal is_tag_r             : std_logic;                                                      --! Current segment is a tag
    signal is_nsec_r            : std_logic;                                                      --! Current block is a nsec
    signal is_ready_r           : std_logic;                                                      --! Current block is ready
    signal is_pad_r             : std_logic;                                                      --! Current block contains padding
    signal is_eot_r             : std_logic;                                                      --! Current block is the last of its type in the current segment
    signal is_eom_r             : std_logic;                                                      --! Current block is the last block of message
    signal is_nodata_r          : std_logic;                                                      --! Current block contains no message data (used for authenticated encryption data only mode)
    signal is_first_blk         : std_logic;
    signal is_init_r            : std_logic;

    --!     Padding related signals
    signal pad_done             : std_logic;

    --!     Other status registers
    signal is_extra_block       : std_logic;
    signal is_partial_data      : std_logic;
    signal total_words          : std_logic_vector(WCOUNT_SIZE                -1 downto 0);

    --! Error registers
    signal errors               : std_logic_vector(8                          -1 downto 0);         --! Error flag (Used for debugging)
    signal error_pdi            : std_logic_vector(8 downto 1);
    signal error_sdi            : std_logic_vector(8 downto 1);
begin
    --! Datapath
    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                key_needs_update_r <= '0';
                key_ready_r      <= '0';
                bdi_ready        <= '0';
            else
                if (set_key_needs_update = '1') then
                    key_needs_update_r <= '1';
                elsif (key_updated = '1') then
                    key_needs_update_r <= '0';
                end if;

                if (set_key_ready = '1') then
                    key_ready_r <= '1';
                elsif (key_updated = '1') then
                    key_ready_r <= '0';
                end if;

                if G_DISABLE_NPUB_PORT = 0 then
                    if set_npub_ready = '1' then
                        npub_ready <= '1';
                    elsif clr_npub_ready = '1' then
                        npub_ready <= '0';
                    end if;
                end if;

                bdi_ready     <= is_ready_r and not bdi_read;
            end if;
        end if;
    end process;

    key_ready           <= key_ready_r;
    key_needs_update    <= key_needs_update_r;
    pad_shift           <= p_size_r(log2_ceil(G_W/8)           -1 downto 0);

    s_opcode            <= sdi(G_SW-12                -1 downto G_SW-16);
    s_stype             <= sdi(G_SW- 8                -1 downto G_SW-12);
    s_size              <= sdi(G_SW-16                -1 downto      0);

    p_size              <= pdi(CNTR_WIDTH             -1 downto      0);
    p_opcode            <= pdi(G_W-12                 -1 downto G_W-16);
    p_stype             <= pdi(G_W- 8                 -1 downto G_W-12);

    bdi_ad              <= is_ad_r;
    bdi_nsec            <= is_nsec_r;
    bdi_decrypt         <= is_decrypt_r;
    bdi_pad             <= is_pad_r;
    bdi_eot             <= is_eot_r;
    bdi_eoi             <= is_eom_r;
    bdi_nodata          <= is_nodata_r;

    errors              <= error_pdi or error_sdi;
    error               <= nway_or(errors);

    genPad: if (G_PAD = 1) generate
        en_pad_loc <= '1' when (pstate = SP_WAIT_DATA and p_lst = '1' and pwrem <= G_W/8 and pad_done = '0') else '0';
    end generate;

    --! Control for public data
    p_pstate:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                pstate          <= SP_WAIT_INSTR;
                en_data         <= '0';
                en_npub           <= '0';
                error_pdi       <= (others => '0');
                bypass_fifo_wr  <= '0';
                p_lst           <= '0';
                p_lsm           <= '0';
                p_pad           <= '0';
                p_id            <= (others => '0');
                bdi_proc        <= '0';
                is_ad_r         <= '0';
                is_tag_r        <= '0';
                is_nsec_r       <= '0';
                is_ae_r         <= '0';
                is_decrypt_r    <= '0';
                is_ready_r      <= '0';
                is_pad_r        <= '0';
                is_eot_r        <= '0';
                is_eom_r        <= '0';
                is_nodata_r     <= '0';
                sel_blank_pdi   <= '0';
                exp_tag_ready   <= '0';
                if PARTIAL_LOAD = 1 then
                    is_partial_data <= '0';
                end if;

                p_reg_stype     <= (others => '0');
                set_key_needs_update    <= '0';

                --! Pad
                if G_PAD = 1 then
                    pad_enable      <= '0';
                end if;
            else
                en_data         <= '0';
                en_npub           <= '0';
                bypass_fifo_wr  <= '0';
                set_npub_ready  <= '0';
                clr_npub_ready  <= '0';
                clr_len         <= '0';
                sel_blank_pdi   <= '0';
                en_exp_tag      <= '0';
                set_key_needs_update    <= '0';
                pad_type_ad     <= '0';
                pad_eot         <= '0';
                en_last_word    <= '0';
                if ((G_DBLK_SIZE MOD G_W) > 0) then
                    sel_input <= (others => '0');
                end if;

                --! Pad
                if G_PAD = 1 then
                    pad_enable      <= '0';
                end if;
                if (G_KEYAK = 1) then
                    key_updated_int <= '0';
                end if;

                case pstate is
                    when SP_WAIT_INSTR =>
                        if (pdi_valid = '1' and bypass_fifo_full = '0') then
                            pstate          <= SP_READ_INSTR;
                            bypass_fifo_wr  <= '1';
                        end if;

                        pwrem        <= (others => '0');
                        p_size_r     <= (others => '0');
                        p_lst        <= '0';
                        p_lsm        <= '0';
                        p_pad        <= '0';
                        p_id         <= (others => '0');
                        wcount       <= (others => '0');
                        is_ad_r      <= '0';
                        is_tag_r     <= '0';
                        is_nsec_r    <= '0';
                        is_ae_r      <= '0';
                        is_decrypt_r <= '0';
                        is_ready_r   <= '0';
                        is_pad_r     <= '0';
                        clr_npub_ready <= '1';
                        clr_len      <= '1';
                        is_nodata_r  <= '0';
                        exp_tag_ready <= '0';
                        is_extra_block <= '0';
                        bdi_size     <= (others => '0');
                        if G_REVERSE_DBLK = 1 then
                            is_first_blk <= '1';
                        end if;
                        if (G_KEYAK = 1) then
                            is_init_r    <= '1';
                        end if;

                    when SP_READ_INSTR =>
                        if  (p_opcode /= OP_AE_ENC) and (p_opcode /= OP_AE_DEC) and (p_opcode /= OP_ACT_KEY) and
                            (p_opcode /= OP_ENC) and (p_opcode /= OP_DEC)
                        then
                            pstate              <= SP_ERROR;
                            error_pdi(1)        <= '1';    --! Unsupported opcode
                        else
                            if (p_opcode = OP_AE_DEC) or (p_opcode = OP_DEC) then
                                is_decrypt_r    <= '1';
                            end if;
                            if (p_opcode = OP_AE_DEC or p_opcode = OP_AE_ENC) then
                                is_ae_r         <= '1';
                            end if;
                            if (p_opcode = OP_ACT_KEY) then
                                pstate                  <= SP_WAIT_INSTR;
                                set_key_needs_update    <= '1';
                            else
                                pstate                  <= SP_WAIT_HDR;
                                p_reg_stype             <= p_stype;
                                bdi_proc                <= '1';
                            end if;
                        end if;

                    when SP_WAIT_HDR =>
                        if (pdi_valid = '1' and bypass_fifo_full = '0') then
                            pstate              <= SP_READ_HDR;
                            bypass_fifo_wr      <= '1';
                        end if;
                        --! Pad
                        if G_PAD = 1 then
                            pad_done            <= '0';
                        end if;

                    when SP_READ_HDR =>
                        if (p_stype /= ST_NPUB) and (p_stype /= ST_AD) and (p_stype /= ST_NSEC) and (p_stype /= ST_NSEC_CIPH) and
                           (p_stype /= ST_MESSAGE and is_decrypt_r = '0') and
                           (p_stype /= ST_CIPHER  and is_decrypt_r = '1') and
                           (p_stype /= ST_TAG     and is_decrypt_r = '1')
                        then
                            pstate       <= SP_ERROR;
                            error_pdi(7) <= '1';                            --! Segment type incompatible with opcode
                        else
                            --! Store segment type
                            if (pdi(G_W-15-1) = '1') then
                                p_reg_stype <= p_stype;
                            end if;
                            --! Determining the type and appropriate sequence of inputs. Uses for debugging as well (not fully implemented).
                            if    (p_stype = ST_NPUB) then
                                if ( p_reg_stype /= ST_INIT ) then
                                    pstate              <= SP_ERROR;
                                    error_pdi(6)        <= '1';             --! Segment out of sequence
                                else
                                    if ( p_size /= G_NPUB_SIZE/8 ) then
                                        pstate          <= SP_ERROR;
                                        error_pdi(4)    <= '1';             --! Unsupported NPUB Length
                                    else
                                        pstate          <= SP_WAIT_NPUB;
                                    end if;
                                end if;
                            elsif (p_stype = ST_NSEC or p_stype = ST_NSEC_CIPH) then
                                if (p_reg_stype /= ST_NPUB) then
                                    pstate          <= SP_ERROR;
                                    error_pdi(6)    <= '1';                 --! Segment out of sequence
                                else
                                    pstate          <= SP_WAIT_DATA;
                                    is_nsec_r       <= '1';
                                end if;
                            elsif (p_stype = ST_AD) then
                                if (G_PLAINTEXT_MODE = 0 and (p_reg_stype = ST_NPUB   or p_reg_stype = ST_AD)) or
                                   (G_PLAINTEXT_MODE = 1 and (p_reg_stype = ST_INIT or p_reg_stype = ST_AD)) or
                                   (G_PLAINTEXT_MODE = 2 and (p_reg_stype = ST_INIT or p_reg_stype = ST_AD)) or
                                   (G_PLAINTEXT_MODE = 3 and (p_reg_stype = ST_NPUB   or p_reg_stype = ST_AD or p_reg_stype = ST_MESSAGE or p_reg_stype = ST_CIPHER)) or
                                   ((G_CIPHERTEXT_MODE = 2 and G_REVERSE_DBLK = 1) and (p_reg_stype = ST_TAG))
                                then
                                    if (G_KEYAK = 1 and is_init_r = '1') then
                                        is_init_r       <= '0';
                                        pstate          <= SP_INIT_KEYAK0;
                                    else
                                        pstate          <= SP_WAIT_DATA;
                                    end if;
                                    is_ad_r         <= '1';
                                    is_tag_r        <= '0';
                                    is_nsec_r       <= '0';
                                else
                                    pstate          <= SP_ERROR;
                                    error_pdi(6)    <= '1';                 --! Segment out of sequence
                                end if;

                                if (PARTIAL_LOAD = 1) then
                                    total_words <= std_logic_vector(to_unsigned(CNT_AD_WORDS, WCOUNT_SIZE));
                                end if;
                            elsif (p_stype = ST_MESSAGE or p_stype = ST_CIPHER) then
                                if (G_PLAINTEXT_MODE = 0 and (p_reg_stype = ST_NPUB   or p_reg_stype = ST_AD or p_reg_stype = ST_MESSAGE or p_reg_stype = ST_CIPHER)) or
                                   (G_PLAINTEXT_MODE = 1) or
                                   (G_PLAINTEXT_MODE = 2) or
                                   (G_PLAINTEXT_MODE = 3 and (p_reg_stype = ST_NPUB   or p_reg_stype = ST_AD or p_reg_stype = ST_MESSAGE or p_reg_stype = ST_CIPHER)) or
                                   ((G_CIPHERTEXT_MODE = 2 and G_REVERSE_DBLK = 1) and (p_reg_stype = ST_TAG))
                                then
                                    pstate          <= SP_WAIT_DATA;
                                    is_ad_r         <= '0';
                                    is_tag_r        <= '0';
                                    is_nsec_r       <= '0';
                                else
                                    pstate          <= SP_ERROR;
                                    error_pdi(6)    <= '1';                 --! Segment out of sequence
                                end if;

                                if (PARTIAL_LOAD = 1) then
                                    total_words <= std_logic_vector(to_unsigned(CNT_DATA_WORDS, WCOUNT_SIZE));
                                end if;

                                --! Calculate if additional block needs to be added (only for Message type)
                                if (G_CIPHERTEXT_MODE = 2 and G_PAD_D = 2) then
                                    if (pdi(G_BS_BYTES-1 downto 0) = 0) then
                                        is_extra_block  <= '1';
                                    else
                                        is_extra_block  <= '0';
                                    end if;
                                end if;

                            elsif (p_stype = ST_TAG) then
                                if (is_decrypt_r  = '0') then
                                    pstate          <= SP_ERROR;
                                    error_pdi(6)    <= '1';
                                else
                                    pstate          <= SP_WAIT_DATA;
                                    is_tag_r        <= '1';
                                    if (is_ready_r = '0') then
                                        is_ad_r         <= '1';                 --! Operate like AD
                                        is_nsec_r       <= '0';
                                    end if;
                                end if;
                            else    --! Unsupported segment type
                                pstate          <= SP_ERROR;
                                error_pdi(6)    <= '1';
                            end if;
                            --! Saves the block state and size to internal register
                            if (G_CIPHERTEXT_MODE = 2 and p_stype = ST_CIPHER) then
                                pwrem           <= (pdi(CNTR_WIDTH-1 downto G_BS_BYTES) + 1) & ZEROS(G_BS_BYTES-1 downto 0);
                            else
                                pwrem           <= pdi(CNTR_WIDTH-1 downto 0);
                            end if;

                            p_size_r            <= pdi(CNTR_WIDTH-1 downto 0);
                            p_lst               <= pdi(G_W-15-1);
                            p_lsm               <= pdi(G_W-14-1);
                            p_pad               <= pdi(G_W-13-1);
                            p_id                <= pdi(G_W-1    downto G_W-8);
                        end if;

                    when SP_WAIT_NPUB =>
                        if (pdi_valid = '1' and bypass_fifo_full = '0') then
                            pwrem       <= pwrem - G_W/8;
                            if G_DISABLE_NPUB_PORT = 1 then
                                en_data     <= '1';
                            else
                                en_npub       <= '1';
                            end if;
                            bypass_fifo_wr  <= '1';
                            if (wcount = CNT_NPUB_WORDS-1) then
                                if p_lsm = '1' or G_DISABLE_NPUB_PORT = 1 then
                                    pstate      <= SP_WAIT_DATA_READ;
                                    is_nodata_r <= '1';
                                else
                                    pstate      <= SP_WAIT_HDR;
                                end if;
                                wcount          <= (others => '0');
                                if G_DISABLE_NPUB_PORT = 1 then
                                    is_ready_r      <= '1';
                                else
                                    set_npub_ready    <= '1';
                                end if;
                            else
                                wcount          <= wcount + 1;
                            end if;
                        end if;

                    when SP_INIT_KEYAK0 =>
                        if (key_needs_update_r = '1' and key_ready_r = '1') then
                            key_updated_int <= '1';
                            pstate          <= SP_INIT_KEYAK1;
                        elsif (key_needs_update_r = '0') then
                            pstate          <= SP_INIT_KEYAK1;
                        end if;

                    when SP_INIT_KEYAK1 =>
                        sel_input <= "100";
                        en_data   <= '1';
                        wcount <= wcount + 1;
                        pstate <= SP_INIT_KEYAK2;

                    when SP_INIT_KEYAK2 =>
                        sel_input <= "101";
                        en_data   <= '1';
                        wcount    <= wcount + 1;
                        pstate    <= SP_WAIT_DATA;

                    when SP_WAIT_DATA =>
                        if ((is_ad_r = '1' or is_nsec_r = '1') and bypass_fifo_full = '0' and pdi_valid = '1') or
                           (is_ad_r = '0' and pdi_valid = '1') or
                           (pwrem = 0     and p_lst = '1')
                        then
                            --! If remaining data word is 0, pad input block with 0
                            if (pwrem /= 0) then
                                if (is_ad_r = '1' or is_tag_r = '1') then
                                    bypass_fifo_wr  <= '1';
                                end if;
                            else
                                sel_blank_pdi       <= '1';
                            end if;
                            --! Perform padding
                            if G_PAD = 1 then
                                if (p_lst = '1' and pwrem < G_W/8 and pad_done = '0') then
                                    if ((is_ad_r = '0' and G_PAD_D > 0) or (is_ad_r = '1' and is_tag_r = '0' and G_PAD_AD > 0) or (G_PAD_D > 0 and G_PAD_AD > 0)) then
                                        pad_enable      <= '1';
                                        is_pad_r        <= '1';
                                        pad_done        <= '1';
                                        pad_eot         <= p_lst;
                                        pad_type_ad     <= is_ad_r;
                                    end if;
                                end if;
                            end if;

                            if (pwrem <= G_W/8) then
                                pwrem               <= (others => '0');
                                size_dword          <= pwrem(log2_ceil(G_W/8) downto 0);
                            else
                                pwrem               <= pwrem - G_W/8;
                                size_dword          <= std_logic_vector(to_unsigned(G_W/8,log2_ceil(G_W/8)+1));
                            end if;


                            if (is_tag_r = '0') then
                                en_data         <= '1';

                                --! Handle the case when ((G_DBLK_SIZE mod G_W) /= 0).
                                --! Note: Only when (G_DBLK_SIZE mod G_W) = (G_W/2).
                                if (PARTIAL_LOAD = 1) then
                                    if (wcount = total_words-1) then
                                        en_last_word    <= '1';
                                        if (pwrem >  (G_W/8)/2) then
                                            is_partial_data <= not is_partial_data;
                                        end if;
                                    elsif (pwrem = 0) then
                                       is_partial_data <= '0';
                                    end if;

                                    if (is_partial_data = '0') then
                                        sel_input <= "000"; --! D[HI] & D[LO]
                                    else
                                        if (pwrem = 0) then
                                            sel_input <= "011"; --! "00..."
                                        elsif (pwrem <= (G_W/8)/2) then
                                            sel_input <= "010"; --! D[LO] & 00..
                                        else
                                            sel_input <= "001"; --! D[LO] & D[HI]
                                        end if;
                                    end if;
                                end if;

                                if (wcount = CNT_DATA_WORDS-1) then
                                    p_size_r <= p_size_r - G_DBLK_SIZE/8;
                                    if ((G_REVERSE_DBLK = 0) or (is_decrypt_r = '0') or (is_ad_r = '1')) then
                                        if (pwrem < G_W/8) then
                                            bdi_size        <= p_size_r(G_BS_BYTES-1 downto 0);
                                        else
                                            bdi_size        <= (others => '0');
                                        end if;
                                    else
                                        if (is_first_blk = '1' and is_ad_r = '0') then
                                            if (remainder(G_BS_BYTES-1 downto 0) < G_DBLK_SIZE/8) then
                                                bdi_size    <= remainder(G_BS_BYTES-1 downto 0);
                                            else
                                                bdi_size    <= (others => '0');
                                            end if;
                                            is_first_blk <= '0';
                                        else
                                            bdi_size        <= (others => '0');
                                        end if;
                                    end if;
                                    wcount <= (others => '0');

                                    --! Status signals and state transition
                                    is_ready_r          <= '1';
                                    if (pwrem <= G_W/8) then
                                        if (G_PAD_D = 2 and is_extra_block = '1') then
                                            is_eot_r        <= '0';
                                        else
                                            is_eot_r        <= p_lst;
                                        end if;

                                        --! If it's decrypt and of type message, start loading tag segment
                                        if (is_decrypt_r = '1' and is_ad_r = '0' and p_lst = '1') then
                                            is_eom_r    <= p_lst;
                                            if (G_REVERSE_DBLK = 1 and G_CIPHERTEXT_MODE = 2) then
                                                if (p_lsm = '1') then
                                                    pstate      <= SP_WAIT_MSG_AUTH;
                                                else
                                                    pstate      <= SP_WAIT_HDR;
                                                end if;
                                            else
                                                pstate      <= SP_WAIT_HDR;
                                            end if;
                                        else
                                            if (G_PAD_D = 2 and is_extra_block = '1') then
                                                is_eom_r        <= '0';
                                            else
                                                is_eom_r        <= p_lsm;
                                            end if;
                                            pstate      <= SP_WAIT_DATA_READ;
                                        end if;
                                    else
                                        pstate <= SP_WAIT_DATA_READ;
                                    end if;

                                    if G_PAD = 0 then   --! Disable this clause if padding is performed internally.
                                        is_pad_r        <= p_pad;
                                    end if;
                                else
                                    wcount <= wcount + 1;
                                end if;
                            else
                                if (bdi_read = '1') then
                                    is_ready_r <= '0';
                                end if;

                                en_exp_tag          <= '1';
                                if (wcount = CNT_TAG_WORDS-1) then
                                    wcount <= (others => '0');
                                    if (p_lsm = '1') then
                                        pstate        <= SP_WAIT_MSG_AUTH;
                                    else
                                        pstate        <= SP_WAIT_HDR;
                                    end if;
                                    exp_tag_ready <= '1';
                                else
                                    wcount <= wcount + 1;
                                end if;
                            end if;
                        end if;

                    when SP_WAIT_DATA_READ =>
                        wcount <= (others => '0');
                        if (bdi_read = '1') then
                            is_ready_r      <= '0';
                            is_eot_r        <= '0';
                            is_eom_r        <= '0';
                            is_pad_r        <= '0';
                            if (pwrem = 0) then
                                if (PARTIAL_LOAD = 1 and is_partial_data = '1') or
                                   (G_PAD_D = 2 and p_reg_stype = ST_MESSAGE and is_extra_block = '1') 
                                then
                                    --! Special I/O size
                                    pstate <= SP_WAIT_DATA;                                    
                                    if (G_PAD_D = 2 and p_reg_stype = ST_MESSAGE and is_extra_block = '1') then
                                        is_extra_block  <= '0';
                                    end if;
                                else
                                    --! Standard I/O size
                                    if (p_lsm = '1') then                 
                                        if (is_decrypt_r = '1') then    --! If AEAD_DECRYPT, there must be a tag segment after last message segment
                                            pstate      <= SP_WAIT_HDR;
                                        else
                                            pstate        <= SP_WAIT_INSTR;
                                            bdi_proc      <= '0';
                                        end if;
                                    else
                                        pstate  <= SP_WAIT_HDR;
                                    end if;
                                end if;
                            else
                                pstate <= SP_WAIT_DATA;
                            end if;
                        end if;

                    when SP_WAIT_MSG_AUTH =>
                        if (msg_auth_done = '1') then
                            pstate        <= SP_WAIT_INSTR;
                            bdi_proc      <= '0';
                            is_ready_r    <= '0';
                            is_eot_r      <= '0';
                            is_eom_r      <= '0';
                            is_pad_r      <= '0';
                        elsif (bdi_read = '1') then
                            is_ready_r <= '0';
                        end if;

                    when SP_ERROR =>
                        pstate      <= SP_ERROR;
                end case;
            end if;
        end if;
    end process;
    --! Unregistered control signals (Required for correct timing)
    pdi_ready         <= '1' when pdi_valid = '1' and (
                        (pstate  = SP_WAIT_INSTR and bypass_fifo_full = '0') or
                        (pstate  = SP_WAIT_HDR   and bypass_fifo_full = '0') or
                        (pstate  = SP_WAIT_NPUB  and bypass_fifo_full = '0') or
                        (pstate  = SP_WAIT_DATA  and (is_ad_r     = '1' or is_nsec_r = '1') and bypass_fifo_full = '0' and pwrem /= 0) or
                        (pstate  = SP_WAIT_DATA  and is_ad_r     = '0' and pwrem /= 0)) else '0';
    en_len_a <= '1' when (pstate = SP_READ_HDR   and p_stype = ST_AD)      else '0';
    en_len_d <= '1' when (pstate = SP_READ_HDR   and (p_stype = ST_MESSAGE or p_stype = ST_CIPHER)) else '0';


    --! Division required for reverse block mode
    gRB:
    if (G_REVERSE_DBLK = 1) generate
        remainder <= std_logic_vector(unsigned(p_size_r) MOD to_unsigned(G_DBLK_SIZE/8, CNTR_WIDTH));
    end generate;



    --! Control for secret data
    p_sstate:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                sstate        <= SS_WAIT_INSTR;
                set_key_ready <= '0';
                error_sdi     <= (others => '0');
            else
                en_key        <= '0';
                set_key_ready <= '0';
                case sstate is
                    when SS_WAIT_INSTR =>
                        if (sdi_valid = '1') then
                            sstate          <= SS_READ_INSTR;
                        end if;
                    when SS_READ_INSTR =>
                        s_reg_stype        <= s_stype;

                        if (s_opcode /= OP_LD_KEY) then
                            sstate          <= SS_WAIT_INSTR;
                            error_sdi(1)    <= '1';                     --! Unsupported opcode
                        else
                            sstate          <= SS_WAIT_HDR;
                        end if;
                    when SS_WAIT_HDR =>
                        if sdi_valid = '1' then
                            sstate  <= SS_READ_HDR;
                        end if;
                    when SS_READ_HDR =>
                        sstate      <= SS_WAIT_DATA;
                        if (s_stype /= ST_KEY) or (s_size /= G_KEY_SIZE/8) then
                            if (s_stype /= ST_KEY) then
                                error_sdi(7)    <= '1';                 --! Segment type incompatible with opcode
                            end if;
                            if (s_size /= G_KEY_SIZE/8) then
                                error_sdi(3)    <= '1';                 --! Unsupported Key Length
                            end if;
                        else
                            swrem <= std_logic_vector(to_unsigned(G_KEY_SIZE/G_SW, LOG2_KEYWORDS));
                        end if;
                    when SS_WAIT_DATA =>
                        if (sdi_valid = '1' and key_ready_r = '0') then
                            en_key          <= '1';
                            if (swrem = 1) then
                                sstate        <= SS_WAIT_INSTR;
                                set_key_ready <= '1';
                            else
                                swrem   <= swrem - 1;
                            end if;
                        end if;
                end case;
            end if;
        end if;
    end process;

    sdi_ready <= '1' when sdi_valid = '1' and (
                            sstate = SS_WAIT_INSTR or
                            sstate = SS_WAIT_HDR   or
                            (sstate = SS_WAIT_DATA and key_ready_r = '0')) else '0';
end behavior;

