-------------------------------------------------------------------------------
--! @file       CipherCore_Wrapper.vhd
--! @brief      Wrapper for CipherCore
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2015 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @version    1.0b1
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;

entity CipherCore_Wrapper is
    generic (
        G_NPUB_SIZE              : integer := 96;   --! NPUB size (bits)
        G_DBLK_SIZE              : integer := 128;  --! Data Block size (bits)
        G_KEY_SIZE               : integer := 128;  --! Key size (bits)
        G_TAG_SIZE               : integer := 128;  --! Tag size (bits)
        G_BS_BYTES               : integer := 4;    --! The number of bits required to hold block size expressed in bytes = log2_ceil(max(G_ABLK_SIZE,G_DBLK_SIZE)/8)
        G_CTR_AD_SIZE            : integer := 64;   --! Maximum size for the counter that keeps track of authenticated data
        G_CTR_D_SIZE             : integer := 64    --! Maximum size for the counter that keeps track of data
    );
    port (
        --! Global signals
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! SERDES signals
        sin             : in  std_logic;
        ssel            : in  std_logic;
        sout            : out std_logic
    );
end entity CipherCore_Wrapper;

architecture structure of CipherCore_Wrapper is
    signal sipo         : std_logic_vector(G_DBLK_SIZE/4+G_BS_BYTES+G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto 0);
    signal piso         : std_logic_vector(G_BS_BYTES+1+6+G_TAG_SIZE+G_DBLK_SIZE-1 downto 0);
    signal piso_data    : std_logic_vector(G_BS_BYTES+1+6+G_TAG_SIZE+G_DBLK_SIZE-1 downto 0);
begin
    pReg:
    process(clk)
    begin
        if rising_edge(clk) then
            sipo <= sin & sipo(G_DBLK_SIZE/4+G_BS_BYTES+G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto 1);
            if (ssel = '1') then
                piso <= piso_data;
            else
                piso <= '0' & piso(G_BS_BYTES+1+6+G_TAG_SIZE+G_DBLK_SIZE-1  downto 1);
            end if;
        end if;
    end process;
    sout <= piso(0);

    uuw:
    entity work.CipherCore(structure)
    generic map (
        G_NPUB_SIZE         => G_NPUB_SIZE  ,
        G_DBLK_SIZE         => G_DBLK_SIZE  ,
        G_KEY_SIZE          => G_KEY_SIZE   ,
        G_TAG_SIZE          => G_TAG_SIZE   ,
        G_BS_BYTES          => G_BS_BYTES   ,
        G_CTR_AD_SIZE       => G_CTR_AD_SIZE,
        G_CTR_D_SIZE        => G_CTR_D_SIZE
    )
    port map (
        clk                 => clk,
        rst                 => rst,
        npub                => sipo(                                                                                         G_NPUB_SIZE-1 downto                                                                                                    0),
        key                 => sipo(                                                                              G_KEY_SIZE+G_NPUB_SIZE-1 downto                                                                                          G_NPUB_SIZE),
        bdi                 => sipo(                                                                  G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto                                                                               G_KEY_SIZE+G_NPUB_SIZE),
        exp_tag             => sipo(                                                       G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto                                                                   G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE),
        key_ready           => sipo(                                                     1+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        key_needs_update    => sipo(                                                     2+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        npub_ready          => sipo(                                                     3+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_ready           => sipo(                                                     4+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_proc            => sipo(                                                     5+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_ad              => sipo(                                                     6+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_nsec            => sipo(                                                     7+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_pad             => sipo(                                                     8+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_decrypt         => sipo(                                                     9+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_eot             => sipo(                                                    10+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_eoi             => sipo(                                                    11+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        exp_tag_ready       => sipo(                                                    12+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdo_ready           => sipo(                                                    13+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        tag_ready           => sipo(                                                    14+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        bdi_nodata          => sipo(                                                    15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1),
        len_a               => sipo(                                      G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto                                                     15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE),
        len_d               => sipo(                         G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto                                       G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE),
        bdi_size            => sipo(              G_BS_BYTES+G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto                          G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE),
        bdi_valid_bytes     => sipo(G_DBLK_SIZE/8+G_BS_BYTES+G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto               G_BS_BYTES+G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE),
        bdi_pad_loc         => sipo(G_DBLK_SIZE/4+G_BS_BYTES+G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE-1 downto G_DBLK_SIZE/8+G_BS_BYTES+G_CTR_D_SIZE+G_CTR_AD_SIZE+15+G_TAG_SIZE+G_DBLK_SIZE+G_KEY_SIZE+G_NPUB_SIZE),

        bdo                 => piso_data(                          G_DBLK_SIZE-1 downto                        0),
        tag                 => piso_data(               G_TAG_SIZE+G_DBLK_SIZE-1 downto              G_DBLK_SIZE),
        bdo_write           => piso_data(             1+G_TAG_SIZE+G_DBLK_SIZE-1),
        bdi_read            => piso_data(             2+G_TAG_SIZE+G_DBLK_SIZE-1),
        key_updated         => piso_data(             3+G_TAG_SIZE+G_DBLK_SIZE-1),
        tag_write           => piso_data(             4+G_TAG_SIZE+G_DBLK_SIZE-1),
        msg_auth_done       => piso_data(             5+G_TAG_SIZE+G_DBLK_SIZE-1),
        msg_auth_valid      => piso_data(             6+G_TAG_SIZE+G_DBLK_SIZE-1),
        bdo_size            => piso_data(G_BS_BYTES+1+6+G_TAG_SIZE+G_DBLK_SIZE-1 downto 6+G_TAG_SIZE+G_DBLK_SIZE)
    );
end structure;